#
#            PySceneDetect: Python-Based Video Scene Detector
#   -------------------------------------------------------------------
#     [  Site:    https://scenedetect.com                           ]
#     [  Docs:    https://scenedetect.com/docs/                     ]
#     [  Github:  https://github.com/Breakthrough/PySceneDetect/    ]
#
# Copyright (C) 2014 Brandon Castellano <http://www.bcastell.com>.
# PySceneDetect is licensed under the BSD 3-Clause License; see the
# included LICENSE file, or visit one of the above pages for details.
#
"""PySceneDetect scenedetect.scene_manager Tests

This file includes unit tests for the scenedetect.scene_manager.SceneManager class,
which applies SceneDetector algorithms on VideoStream backends.
"""

import pytest

from scenedetect.backends.opencv import VideoStreamCv2
from scenedetect.common import FrameTimecode
from scenedetect.detectors import AdaptiveDetector, ContentDetector
from scenedetect.scene_manager import SceneManager, expand_scenes_to_bounds

TEST_VIDEO_START_FRAMES_ACTUAL = [150, 180, 394]


def test_scene_list(test_video_file):
    """Test SceneManager get_scene_list method with VideoStreamCv2/ContentDetector."""
    video = VideoStreamCv2(test_video_file)
    sm = SceneManager()
    sm.add_detector(ContentDetector())

    video_fps = video.frame_rate
    start_time = FrameTimecode("00:00:05", video_fps)
    end_time = FrameTimecode("00:00:10", video_fps)

    assert end_time.frame_num > start_time.frame_num

    video.seek(start_time)
    sm.auto_downscale = True

    num_frames = sm.detect_scenes(video=video, end_time=end_time)

    assert num_frames == (end_time.frame_num - start_time.frame_num)

    scene_list = sm.get_scene_list()
    assert scene_list
    # Each scene is in the format (Start Timecode, End Timecode)
    assert len(scene_list[0]) == 2

    # First scene should start at start_time and last scene should end at end_time.
    assert scene_list[0][0] == start_time
    assert scene_list[-1][1] == end_time

    for i, _ in enumerate(scene_list):
        assert scene_list[i][0].frame_num < scene_list[i][1].frame_num
        if i > 0:
            # Ensure frame list is sorted (i.e. end time frame of
            # one scene is equal to the start time of the next).
            assert scene_list[i - 1][1] == scene_list[i][0]


def test_get_scene_list_start_in_scene(test_video_file):
    """Test SceneManager `get_scene_list()` method with the `start_in_scene` flag."""
    video = VideoStreamCv2(test_video_file)
    sm = SceneManager()
    sm.add_detector(ContentDetector())

    video_fps = video.frame_rate
    # End time must be short enough that we won't detect any scenes.
    end_time = FrameTimecode(25, video_fps)
    sm.auto_downscale = True
    sm.detect_scenes(video=video, end_time=end_time)
    # Should be an empty list.
    assert len(sm.get_scene_list()) == 0
    # Should be a list with a single element spanning the video duration.
    scene_list = sm.get_scene_list(start_in_scene=True)
    assert len(scene_list) == 1
    assert scene_list[0][0] == 0
    assert scene_list[0][1] == end_time


# TODO: This would be more readable if the callbacks were defined within the test case, e.g.
# split up the callback function and callback lambda test cases.
class FakeCallback:
    """Fake callback used for testing. Tracks the frame numbers the callback was invoked with."""

    def __init__(self):
        self.scene_list: list[int] = []

    def get_callback_lambda(self):
        """For testing using a lambda.."""
        return lambda image, frame_num: self._callback(image, frame_num)

    def get_callback_func(self):
        """For testing using a callback function."""

        def callback(image, frame_num):
            nonlocal self
            self._callback(image, frame_num)

        return callback

    def _callback(self, image, frame_num):
        self.scene_list.append(frame_num)


def test_detect_scenes_callback(test_video_file):
    """Test SceneManager detect_scenes method with a callback function.

    Note that the API signature of the callback will undergo breaking changes in v1.0.
    """
    video = VideoStreamCv2(test_video_file)
    sm = SceneManager()
    sm.add_detector(ContentDetector())

    fake_callback = FakeCallback()

    video_fps = video.frame_rate
    start_time = FrameTimecode("00:00:05", video_fps)
    end_time = FrameTimecode("00:00:15", video_fps)
    video.seek(start_time)
    sm.auto_downscale = True

    _ = sm.detect_scenes(
        video=video, end_time=end_time, callback=fake_callback.get_callback_lambda()
    )
    scene_list = sm.get_scene_list()
    assert [start for start, end in scene_list] == TEST_VIDEO_START_FRAMES_ACTUAL
    assert fake_callback.scene_list == TEST_VIDEO_START_FRAMES_ACTUAL[1:]

    # Perform same test using callback function instead of lambda.
    sm.clear()
    sm.add_detector(ContentDetector())
    fake_callback = FakeCallback()
    video.seek(start_time)

    _ = sm.detect_scenes(video=video, end_time=end_time, callback=fake_callback.get_callback_func())
    scene_list = sm.get_scene_list()
    assert [start for start, end in scene_list] == TEST_VIDEO_START_FRAMES_ACTUAL
    assert fake_callback.scene_list == TEST_VIDEO_START_FRAMES_ACTUAL[1:]


def test_detect_scenes_callback_adaptive(test_video_file):
    """Test SceneManager detect_scenes method with a callback function and a detector which
    requires frame buffering.

    Note that the API signature of the callback will undergo breaking changes in v1.0.
    """
    video = VideoStreamCv2(test_video_file)
    sm = SceneManager()
    sm.add_detector(AdaptiveDetector())

    fake_callback = FakeCallback()

    video_fps = video.frame_rate
    start_time = FrameTimecode("00:00:05", video_fps)
    end_time = FrameTimecode("00:00:15", video_fps)
    video.seek(start_time)
    sm.auto_downscale = True

    _ = sm.detect_scenes(
        video=video, end_time=end_time, callback=fake_callback.get_callback_lambda()
    )
    scene_list = sm.get_scene_list()
    assert [start for start, end in scene_list] == TEST_VIDEO_START_FRAMES_ACTUAL
    assert fake_callback.scene_list == TEST_VIDEO_START_FRAMES_ACTUAL[1:]

    # Perform same test using callback function instead of lambda.
    sm.clear()
    sm.add_detector(AdaptiveDetector())
    fake_callback = FakeCallback()
    video.seek(start_time)

    _ = sm.detect_scenes(video=video, end_time=end_time, callback=fake_callback.get_callback_func())
    scene_list = sm.get_scene_list()
    assert [start for start, end in scene_list] == TEST_VIDEO_START_FRAMES_ACTUAL
    assert fake_callback.scene_list == TEST_VIDEO_START_FRAMES_ACTUAL[1:]


def test_detect_scenes_crop(test_video_file):
    video = VideoStreamCv2(test_video_file)
    sm = SceneManager()
    sm.crop = (10, 10, 1900, 1000)
    sm.add_detector(ContentDetector())

    video_fps = video.frame_rate
    start_time = FrameTimecode("00:00:05", video_fps)
    end_time = FrameTimecode("00:00:15", video_fps)
    video.seek(start_time)
    sm.auto_downscale = True

    _ = sm.detect_scenes(video=video, end_time=end_time)
    scene_list = sm.get_scene_list()
    assert [start for start, _ in scene_list] == TEST_VIDEO_START_FRAMES_ACTUAL


def test_crop_invalid():
    sm = SceneManager()
    sm.crop = None  # type: ignore[assignment]
    sm.crop = (0, 0, 0, 0)
    sm.crop = (1, 1, 0, 0)
    sm.crop = (0, 0, 1, 1)
    with pytest.raises(TypeError):
        sm.crop = 1  # type: ignore[assignment]
    with pytest.raises(TypeError):
        sm.crop = (1, 1)  # type: ignore[assignment]
    with pytest.raises(TypeError):
        sm.crop = (1, 1, 1)  # type: ignore[assignment]
    with pytest.raises(ValueError):
        sm.crop = (1, 1, 1, -1)


def test_expand_scenes_to_bounds_two_scenes():
    """Scenes detected inside a sub-window should be extended outward."""
    fps = 10.0
    t0 = FrameTimecode(0, fps)
    t130 = FrameTimecode(130, fps)
    t150 = FrameTimecode(150, fps)
    t170 = FrameTimecode(170, fps)
    t300 = FrameTimecode(300, fps)

    scenes = [(t130, t150), (t150, t170)]
    expanded = expand_scenes_to_bounds(scenes, start=t0, end=t300)

    assert expanded == [(t0, t150), (t150, t300)]


def test_expand_scenes_to_bounds_empty():
    """Empty scene lists pass through unchanged."""
    fps = 10.0
    assert expand_scenes_to_bounds([], FrameTimecode(0, fps), FrameTimecode(100, fps)) == []


def test_expand_scenes_to_bounds_single_scene():
    """A single scene gets both endpoints extended."""
    fps = 10.0
    t0 = FrameTimecode(0, fps)
    t130 = FrameTimecode(130, fps)
    t170 = FrameTimecode(170, fps)
    t300 = FrameTimecode(300, fps)

    scenes = [(t130, t170)]
    expanded = expand_scenes_to_bounds(scenes, start=t0, end=t300)

    assert expanded == [(t0, t300)]


def test_expand_scenes_to_bounds_does_not_mutate_input():
    """The input scene list must not be modified in place."""
    fps = 10.0
    t0 = FrameTimecode(0, fps)
    t130 = FrameTimecode(130, fps)
    t150 = FrameTimecode(150, fps)
    t170 = FrameTimecode(170, fps)
    t300 = FrameTimecode(300, fps)

    scenes = [(t130, t150), (t150, t170)]
    original = list(scenes)
    expand_scenes_to_bounds(scenes, start=t0, end=t300)

    assert scenes == original



#
#            PySceneDetect: Python-Based Video Scene Detector
#   -------------------------------------------------------------------
#     [  Site:    https://scenedetect.com                           ]
#     [  Docs:    https://scenedetect.com/docs/                     ]
#     [  Github:  https://github.com/Breakthrough/PySceneDetect/    ]
#
# Copyright (C) 2023 Brandon Castellano <http://www.bcastell.com>.
# PySceneDetect is licensed under the BSD 3-Clause License; see the
# included LICENSE file, or visit one of the above pages for details.
#
"""Logic for the PySceneDetect command."""

import csv
import logging
import os
import time
import warnings

from scenedetect._cli.context import CliContext
from scenedetect.backends import VideoStreamCv2, VideoStreamMoviePy
from scenedetect.common import FrameTimecode
from scenedetect.platform import get_and_create_path
from scenedetect.scene_manager import CutList, SceneList, get_scenes_from_cuts
from scenedetect.video_stream import SeekError

logger = logging.getLogger("pyscenedetect")


def run_scenedetect(context: CliContext):
    """Perform main CLI application control logic. Run once all command-line options and
    configuration file options have been validated.

    Arguments:
        context: Prevalidated command-line option context to use for processing.
    """
    # No input may have been specified depending on the commands/args that were used.
    logger.debug("Running controller.")
    if context.scene_manager is None:
        logger.debug("No input specified.")
        return

    # Suppress warnings when reading past EOF in MoviePy (#461).
    if VideoStreamMoviePy and isinstance(context.video_stream, VideoStreamMoviePy):
        is_debug = context.config.get_value("global", "verbosity") != "debug"
        if not is_debug:
            warnings.filterwarnings("ignore", module="moviepy")

    if context.load_scenes_input:
        # Skip detection if load-scenes was used.
        logger.info("Skipping detection, loading scenes from: %s", context.load_scenes_input)
        if context.stats_file_path:
            logger.warning("WARNING: -s/--stats will be ignored due to load-scenes.")
        scenes, cuts = _load_scenes(context)
        scenes = _postprocess_scene_list(context, scenes)
        logger.info("Loaded %d scenes.", len(scenes))
    else:
        # Perform scene detection on input.
        result = _detect(context)
        if result is None:
            return
        scenes, cuts = result
        scenes = _postprocess_scene_list(context, scenes)
        # Handle -s/--stats option.
        _save_stats(context)
        if scenes:
            logger.info(
                "Detected %d scenes, average shot length %.1f seconds.",
                len(scenes),
                sum([(end_time - start_time).seconds for start_time, end_time in scenes])
                / float(len(scenes)),
            )
        else:
            logger.info("No scenes detected.")

    # Handle post-processing commands the user wants to run (see scenedetect._cli.commands).
    for handler, kwargs in context.commands:
        handler(context=context, scenes=scenes, cuts=cuts, **kwargs)


def _postprocess_scene_list(context: CliContext, scene_list: SceneList) -> SceneList:
    # Handle --merge-last-scene. If set, when the last scene is shorter than --min-scene-len,
    # it will be merged with the previous one.
    if (
        context.merge_last_scene
        and context.min_scene_len is not None
        and context.min_scene_len > 0
        and len(scene_list) > 1
        and (scene_list[-1][1] - scene_list[-1][0]) < context.min_scene_len
    ):
        new_last_scene = (scene_list[-2][0], scene_list[-1][1])
        scene_list = [*scene_list[:-2], new_last_scene]

    # Handle --drop-short-scenes.
    if (
        context.drop_short_scenes
        and context.min_scene_len is not None
        and context.min_scene_len > 0
    ):
        scene_list = [s for s in scene_list if (s[1] - s[0]) >= context.min_scene_len]

    return scene_list


def _detect(context: CliContext) -> tuple[SceneList, CutList] | None:
    perf_start_time = time.time()
    assert context.scene_manager is not None
    assert context.video_stream is not None
    assert context.frame_skip is not None

    context.ensure_detector()
    if context.start_time is not None:
        logger.debug("Seeking to start time...")
        try:
            context.video_stream.seek(target=context.start_time)
        except SeekError as ex:
            logger.critical(
                "Failed to seek to %s / frame %d: %s",
                context.start_time.get_timecode(),
                context.start_time.frame_num,
                str(ex),
            )
            return None

    num_frames = context.scene_manager.detect_scenes(
        video=context.video_stream,
        duration=context.duration,
        end_time=context.end_time,
        frame_skip=context.frame_skip,
        show_progress=not context.quiet_mode,
    )

    # Handle case where video failure is most likely due to multiple audio tracks (#179).
    # TODO(https://scenedetect.com/issues/380): Ensure this does not erroneusly fire.
    if num_frames <= 0 and isinstance(context.video_stream, VideoStreamCv2):
        logger.critical(
            "Failed to read any frames from video file. This could be caused by the video"
            " having multiple audio tracks. If so, try installing the PyAV backend:\n"
            "      pip install av\n"
            "Or remove the audio tracks by running either:\n"
            "      ffmpeg -i input.mp4 -c copy -an output.mp4\n"
            "      mkvmerge -o output.mkv input.mp4\n"
            "For details, see https://scenedetect.com/faq/"
        )
        return None

    perf_duration = time.time() - perf_start_time
    logger.info(
        "Processed %d frames in %.1f seconds (average %.2f FPS).",
        num_frames,
        perf_duration,
        float(num_frames) / perf_duration,
    )

    # Get list of detected cuts/scenes from the SceneManager to generate the required output
    # files, based on the given commands (list-scenes, split-video, save-images, etc...).
    cut_list = context.scene_manager.get_cut_list(show_warning=False)
    scene_list = context.scene_manager.get_scene_list(start_in_scene=True)

    return scene_list, cut_list


def _save_stats(context: CliContext) -> None:
    """Handles saving the statsfile if -s/--stats was specified."""
    if not context.stats_file_path:
        return
    assert context.stats_manager is not None
    if context.stats_manager.is_save_required():
        path = get_and_create_path(context.stats_file_path, context.output)
        logger.info("Saving frame metrics to stats file: %s", path)
        with open(path, mode="w") as file:
            context.stats_manager.save_to_csv(csv_file=file)
    else:
        logger.debug("No frame metrics updated, skipping update of the stats file.")


def _load_scenes(context: CliContext) -> tuple[SceneList, CutList]:
    assert context.load_scenes_input
    assert context.load_scenes_column_name is not None
    assert context.video_stream is not None
    assert os.path.exists(context.load_scenes_input)

    video_stream = context.video_stream
    with open(context.load_scenes_input) as input_file:
        file_reader = csv.reader(input_file)
        csv_headers = next(file_reader)
        if context.load_scenes_column_name not in csv_headers:
            csv_headers = next(file_reader)
        # Check to make sure column headers are present and then load the data.
        if context.load_scenes_column_name not in csv_headers:
            raise ValueError("specified column header for scene start is not present")
        col_idx = csv_headers.index(context.load_scenes_column_name)

        def calculate_timecode(value: str) -> FrameTimecode:
            # Assume other columns are in seconds except frame numbers.
            if value.isdigit():
                # Frame numbers start from index 1 in the CLI output so we correct for that.
                return FrameTimecode(int(value) - 1, fps=video_stream.frame_rate)
            return FrameTimecode(value, fps=video_stream.frame_rate)

        cut_list = sorted(calculate_timecode(row[col_idx]) for row in file_reader)
        # `SceneDetector` works on cuts, so we have to skip the first scene and place the first
        # cut point where the next scenes starts.
        if cut_list:
            cut_list = cut_list[1:]

        start_time = context.video_stream.base_timecode
        if context.start_time is not None:
            start_time = context.start_time
            cut_list = [cut for cut in cut_list if cut > context.start_time]

        video_duration = context.video_stream.duration
        assert video_duration is not None
        end_time = video_duration
        if context.end_time is not None:
            end_time = min(context.end_time, video_duration)
        elif context.duration is not None:
            end_time = min(start_time + context.duration, video_duration)

        cut_list = [cut for cut in cut_list if cut < end_time]
        scene_list = get_scenes_from_cuts(cut_list=cut_list, start_pos=start_time, end_pos=end_time)

        return (scene_list, cut_list)



#
#            PySceneDetect: Python-Based Video Scene Detector
#   -------------------------------------------------------------------
#     [  Site:    https://scenedetect.com                           ]
#     [  Docs:    https://scenedetect.com/docs/                     ]
#     [  Github:  https://github.com/Breakthrough/PySceneDetect/    ]
#
# Copyright (C) 2026 Brandon Castellano <http://www.bcastell.com>.
# PySceneDetect is licensed under the BSD 3-Clause License; see the
# included LICENSE file, or visit one of the above pages for details.
#
"""PySceneDetect scenedetect.backends.concat Tests

Validates the multi-video concatenation logic in `scenedetect.backends.concat`."""

import pytest

from scenedetect import SceneManager, ThresholdDetector, open_video
from scenedetect.backends import AVAILABLE_BACKENDS
from scenedetect.backends.concat import VideoStreamConcat
from scenedetect.video_stream import VideoOpenFailure

FADES_TOTAL_FRAMES = 250
FADES_DURATION = 10.0

BACKENDS = [backend for backend in ("opencv", "pyav") if backend in AVAILABLE_BACKENDS]


@pytest.mark.parametrize("backend", BACKENDS)
def test_decode_single(test_fades_clip, backend):
    """Decode a single video and validate the reported frame count and position."""
    video = VideoStreamConcat([test_fades_clip], backend=backend)
    while video.read(decode=False) is not False:
        pass
    assert video.frame_number == FADES_TOTAL_FRAMES
    assert video.decode_failures == 0


@pytest.mark.parametrize("backend", BACKENDS)
def test_decode_multiple(test_fades_clip, backend):
    """Decode multiple videos and validate the reported frame count."""
    splice_amount = 3
    video = VideoStreamConcat([test_fades_clip] * splice_amount, backend=backend)
    while video.read(decode=False) is not False:
        pass
    assert video.frame_number == FADES_TOTAL_FRAMES * splice_amount
    assert video.decode_failures == 0


@pytest.mark.parametrize("backend", BACKENDS)
def test_seam_monotonicity(test_fades_clip, backend):
    """Position must be strictly increasing across the file seam."""
    video = VideoStreamConcat([test_fades_clip] * 2, backend=backend)
    last_seconds = -1.0
    max_delta = 0.0
    while video.read(decode=False) is not False:
        seconds = video.position.seconds
        assert seconds > last_seconds, f"position went backwards: {seconds} <= {last_seconds}"
        if last_seconds >= 0:
            max_delta = max(max_delta, seconds - last_seconds)
        last_seconds = seconds
    # The seam should be continuous: no gap larger than a few frame durations.
    assert max_delta < 0.5, f"discontinuity across seam: {max_delta}s"
    assert last_seconds > 2 * FADES_DURATION - 1.0


@pytest.mark.parametrize("backend", BACKENDS)
def test_seek(test_fades_clip, backend):
    """Seeking should work on the global timeline, in either direction, across sources."""
    video = VideoStreamConcat([test_fades_clip] * 2, backend=backend)
    # Seek into the second source.
    target = FADES_DURATION + 5.0
    video.seek(target)
    assert video.read(decode=False) is not False
    assert abs(video.position.seconds - target) < 0.25
    # Seek backwards into the first source.
    video.seek(5.0)
    assert video.read(decode=False) is not False
    assert abs(video.position.seconds - 5.0) < 0.25


def test_seek_backward_then_cross_seam(test_fades_clip):
    """Crossing the seam a second time after a backward seek must not shift the timeline
    again (offset correction must be idempotent)."""
    video = VideoStreamConcat([test_fades_clip] * 2)
    # Read across the seam once.
    video.seek(FADES_DURATION - 0.5)
    while video.position.seconds < FADES_DURATION + 0.5:
        assert video.read(decode=False) is not False
    first_pass = video.position.seconds
    # Seek backward before the seam and cross it again.
    video.seek(FADES_DURATION - 0.5)
    last = video.position.seconds
    while video.position.seconds < FADES_DURATION + 0.5:
        assert video.read(decode=False) is not False
        assert video.position.seconds > last
        last = video.position.seconds
    assert abs(video.position.seconds - first_pass) < 0.25


def test_seam_monotonicity_vfr(test_vfr_drop3_video):
    """Position must also be strictly increasing across the seam between variable framerate
    inputs, whose declared duration is less exact than CFR."""
    video = VideoStreamConcat([test_vfr_drop3_video] * 2)
    last_seconds = -1.0
    while video.read(decode=False) is not False:
        seconds = video.position.seconds
        assert seconds > last_seconds, f"position went backwards: {seconds} <= {last_seconds}"
        last_seconds = seconds


def test_map_span(test_fades_clip):
    """A span crossing the seam between two inputs must map to two local spans."""
    video = VideoStreamConcat([test_fades_clip] * 2)
    duration = FADES_DURATION
    start = video.base_timecode + (duration - 3.0)
    end = video.base_timecode + (duration + 3.0)
    spans = video.map_span(start, end)
    assert len(spans) == 2
    assert spans[0].source_index == 0 and spans[1].source_index == 1
    assert abs(spans[0].local_start.seconds - (duration - 3.0)) < 0.01
    assert abs(spans[0].local_end.seconds - duration) < 0.01
    assert spans[1].local_start.seconds == 0.0
    assert abs(spans[1].local_end.seconds - 3.0) < 0.01
    # A span entirely within the first source maps to a single span.
    spans = video.map_span(video.base_timecode + 1.0, video.base_timecode + 2.0)
    assert len(spans) == 1 and spans[0].source_index == 0


def test_mismatched_resolution(test_fades_clip, test_video_file):
    """Sources with different resolutions cannot be concatenated."""
    with pytest.raises(VideoOpenFailure):
        VideoStreamConcat([test_fades_clip, test_video_file])


def test_unknown_backend_falls_back(test_fades_clip):
    """An unknown backend name falls back to OpenCV instead of failing."""
    video = VideoStreamConcat([test_fades_clip], backend="not_a_backend")
    assert video.child_backend == "opencv"
    assert video.read(decode=False) is not False


def test_open_video_list(test_fades_clip):
    """`open_video` accepts a list of paths and returns a concatenated stream."""
    video = open_video([test_fades_clip, test_fades_clip])
    assert isinstance(video, VideoStreamConcat)
    assert video.duration.seconds == pytest.approx(2 * FADES_DURATION, abs=0.1)
    # A single-element list also returns a concatenated stream.
    video = open_video([test_fades_clip])
    assert isinstance(video, VideoStreamConcat)


def test_scene_manager_detect(test_fades_clip):
    """The concatenated stream must work end-to-end with SceneManager: detecting fades over
    two spliced copies must find twice as many scenes as a single copy."""

    def detect_scenes(paths):
        scene_manager = SceneManager()
        scene_manager.add_detector(ThresholdDetector())
        video = open_video(paths)
        scene_manager.detect_scenes(video=video)
        return scene_manager.get_scene_list()

    single = detect_scenes([test_fades_clip])
    double = detect_scenes([test_fades_clip] * 2)
    assert len(single) > 0
    assert len(double) == 2 * len(single)


@pytest.mark.skipif("pyav" not in AVAILABLE_BACKENDS, reason="PyAV backend not available")
def test_corrupt_concat(corrupt_video_file):
    """The PyAV input path must tolerate corrupt frames and decode the full stream."""
    video = VideoStreamConcat([corrupt_video_file], backend="pyav")
    num_frames = 0
    while video.read(decode=False) is not False:
        num_frames += 1
    assert num_frames >= 590



#
#            PySceneDetect: Python-Based Video Scene Detector
#   -------------------------------------------------------------------
#     [  Site:    https://scenedetect.com                           ]
#     [  Docs:    https://scenedetect.com/docs/                     ]
#     [  Github:  https://github.com/Breakthrough/PySceneDetect/    ]
#
# Copyright (C) 2024 Brandon Castellano <http://www.bcastell.com>.
# PySceneDetect is licensed under the BSD 3-Clause License; see the
# included LICENSE file, or visit one of the above pages for details.
#
"""Logic for PySceneDetect commands that operate on the result of the processing pipeline.

In addition to the the arguments registered with the command, commands will be called with the
current command-line context, as well as the processing result (scenes and cuts).
"""

import logging
import webbrowser
from string import Template

from scenedetect._cli.config import FcpFormat
from scenedetect._cli.context import CliContext
from scenedetect.output import save_images as save_images_impl
from scenedetect.output import (
    split_video_ffmpeg,
    split_video_mkvmerge,
    write_scene_list,
    write_scene_list_edl,
    write_scene_list_fcp7,
    write_scene_list_fcpx,
    write_scene_list_html,
    write_scene_list_otio,
)
from scenedetect.platform import get_and_create_path
from scenedetect.scene_manager import (
    CutList,
    Interpolation,
    SceneList,
    expand_scenes_to_bounds,
)

logger = logging.getLogger("pyscenedetect")


def save_html(
    context: CliContext,
    scenes: SceneList,
    cuts: CutList,
    image_width: int,
    image_height: int,
    filename: str,
    no_images: bool,
    show: bool,
):
    """Handles the `save-html` command."""
    assert context.video_stream is not None
    (image_filenames, output) = (
        context.save_images_result
        if context.save_images_result is not None
        else (None, context.output)
    )

    html_filename = Template(filename).safe_substitute(VIDEO_NAME=context.video_stream.name)
    if not html_filename.lower().endswith(".html"):
        html_filename += ".html"
    html_path = get_and_create_path(html_filename, output)
    write_scene_list_html(
        output_html_filename=html_path,
        scene_list=scenes,
        cut_list=cuts,
        image_filenames=None if no_images else image_filenames,
        image_width=image_width,
        image_height=image_height,
    )
    if show:
        webbrowser.open(html_path)


def save_qp(
    context: CliContext,
    scenes: SceneList,
    cuts: CutList,
    output: str,
    filename: str,
    disable_shift: bool,
):
    """Handler for the `save-qp` command."""
    del scenes  # We only use cuts for this handler.
    assert context.video_stream is not None
    qp_path = get_and_create_path(
        Template(filename).safe_substitute(VIDEO_NAME=context.video_stream.name),
        output,
    )
    start_frame = context.start_time.frame_num if context.start_time else 0
    shift_start = not disable_shift
    offset = start_frame if shift_start else 0
    with open(qp_path, "w") as qp_file:
        qp_file.write(f"{0 if shift_start else start_frame} I -1\n")
        # Place another I frame at each detected cut.
        qp_file.writelines(f"{cut.frame_num - offset} I -1\n" for cut in cuts)
    logger.info(f"QP file written to: {qp_path}")


def list_scenes(
    context: CliContext,
    scenes: SceneList,
    cuts: CutList,
    no_output_file: bool,
    filename: str,
    output: str,
    skip_cuts: bool,
    quiet: bool,
    display_scenes: bool,
    display_cuts: bool,
    cut_format: str,
    col_separator: str,
    row_separator: str,
):
    """Handles the `list-scenes` command."""
    assert context.video_stream is not None
    # Write scene list CSV to if required.
    if not no_output_file:
        scene_list_filename = Template(filename).safe_substitute(
            VIDEO_NAME=context.video_stream.name
        )
        if not scene_list_filename.lower().endswith(".csv"):
            scene_list_filename += ".csv"
        scene_list_path = get_and_create_path(
            scene_list_filename,
            output,
        )
        logger.info("Writing scene list to CSV file:\n  %s", scene_list_path)
        with open(scene_list_path, "w") as scene_list_file:
            write_scene_list(
                output_csv_file=scene_list_file,
                scene_list=scenes,
                include_cut_list=not skip_cuts,
                cut_list=cuts,
                col_separator=col_separator,
                row_separator=row_separator,
            )
    # Suppress output if requested.
    if quiet:
        return
    # Print scene list.
    if display_scenes:
        logger.info(
            """Scene List:
-----------------------------------------------------------------------
 | Scene # | Start Frame |  Start Time  |  End Frame  |   End Time   |
-----------------------------------------------------------------------
%s
-----------------------------------------------------------------------""",
            "\n".join(
                [
                    f" |  {i + 1:5d}  | {start_time.frame_num + 1:11d} | {start_time.get_timecode()} | {end_time.frame_num:11d} | {end_time.get_timecode()} |"
                    for i, (start_time, end_time) in enumerate(scenes)
                ]
            ),
        )
    # Print cut list.
    if cuts and display_cuts:
        logger.info(
            "Comma-separated timecode list:\n  %s",
            ",".join([cut_format.format(cut) for cut in cuts]),
        )


def save_images(
    context: CliContext,
    scenes: SceneList,
    cuts: CutList,
    num_images: int,
    frame_margin: int,
    image_extension: str,
    encoder_param: int,
    filename: str,
    output: str | None,
    show_progress: bool,
    scale: int,
    height: int,
    width: int,
    interpolation: Interpolation,
    threading: bool,
):
    """Handles the `save-images` command."""
    del cuts  # save-images only uses scenes.
    assert context.video_stream is not None

    images = save_images_impl(
        scene_list=scenes,
        video=context.video_stream,
        num_images=num_images,
        frame_margin=frame_margin,
        image_extension=image_extension,
        encoder_param=encoder_param,
        image_name_template=filename,
        output_dir=output,
        show_progress=show_progress,
        scale=scale,
        height=height,
        width=width,
        interpolation=interpolation,
        threading=threading,
    )
    # Save the result for use by `save-html` if required.
    context.save_images_result = (images, output)


def split_video(
    context: CliContext,
    scenes: SceneList,
    cuts: CutList,
    name_format: str,
    use_mkvmerge: bool,
    output: str,
    show_output: bool,
    ffmpeg_args: str,
    expand: bool,
):
    """Handles the `split-video` command."""
    del cuts  # split-video only uses scenes.
    assert context.video_stream is not None

    if expand and scenes:
        video_duration = context.video_stream.duration
        if video_duration is None:
            logger.warning("Cannot --expand: video duration is unavailable for this stream.")
        else:
            scenes = expand_scenes_to_bounds(
                scenes,
                start=context.video_stream.base_timecode,
                end=video_duration,
            )

    if use_mkvmerge:
        name_format = name_format.removesuffix("-$SCENE_NUMBER")

    # Add proper extension to filename template if required.
    dot_pos = name_format.rfind(".")
    extension_length = 0 if dot_pos < 0 else len(name_format) - (dot_pos + 1)
    # If using mkvmerge, force extension to .mkv.
    if use_mkvmerge and not name_format.endswith(".mkv"):
        name_format += ".mkv"
    # Otherwise, if using ffmpeg, only add an extension if one doesn't exist.
    elif not 2 <= extension_length <= 4:
        name_format += ".mp4"
    if use_mkvmerge:
        split_video_mkvmerge(
            input_video_path=context.video_stream.path,
            scene_list=scenes,
            output_dir=output,
            output_file_template=name_format,
            show_output=show_output,
        )
    else:
        split_video_ffmpeg(
            input_video_path=context.video_stream.path,
            scene_list=scenes,
            output_dir=output,
            output_file_template=name_format,
            arg_override=ffmpeg_args,
            show_progress=not context.quiet_mode,
            show_output=show_output,
        )
    if scenes:
        logger.info("Video splitting completed, scenes written to disk.")


def save_edl(
    context: CliContext,
    scenes: SceneList,
    cuts: CutList,
    filename: str,
    output: str,
    title: str,
    reel: str,
    start_timecode: str | None,
):
    """Handles the `save-edl` command. Outputs in CMX 3600 format."""
    del cuts  # We only use scene information.
    assert context.video_stream is not None
    video_name = context.video_stream.name
    edl_path = get_and_create_path(
        Template(filename).safe_substitute(VIDEO_NAME=video_name),
        output,
    )
    write_scene_list_edl(
        output_path=edl_path,
        scene_list=scenes,
        title=Template(title).safe_substitute(VIDEO_NAME=video_name),
        reel=reel,
        start_timecode=start_timecode,
    )


def save_fcp(
    context: CliContext,
    scenes: SceneList,
    cuts: CutList,
    filename: str,
    format: FcpFormat,
    output: str,
):
    """Handles the `save-fcp` command."""
    del cuts  # We only use scene information.
    if not scenes:
        return
    assert context.video_stream is not None

    video_stream = context.video_stream
    video_name = str(video_stream.name)
    video_path = str(video_stream.path)
    xml_path = get_and_create_path(
        Template(filename).safe_substitute(VIDEO_NAME=video_name),
        output,
    )
    if format == FcpFormat.FCPX:
        write_scene_list_fcpx(
            output_path=xml_path,
            scene_list=scenes,
            video_path=video_path,
            frame_rate=video_stream.frame_rate,
            frame_size=video_stream.frame_size,
            video_name=video_name,
        )
    elif format == FcpFormat.FCP7:
        write_scene_list_fcp7(
            output_path=xml_path,
            scene_list=scenes,
            video_path=video_path,
            frame_rate=video_stream.frame_rate,
            frame_size=video_stream.frame_size,
            video_name=video_name,
            source_duration=video_stream.duration,
        )
    else:
        logger.error(f"Unknown format: {format}")


def save_otio(
    context: CliContext,
    scenes: SceneList,
    cuts: CutList,
    filename: str,
    output: str,
    name: str,
    audio: bool,
):
    """Handles the `save-otio` command."""
    del cuts  # We only use scene information
    assert context.video_stream is not None
    video_stream = context.video_stream
    video_name = str(video_stream.name)
    otio_path = get_and_create_path(
        Template(filename).safe_substitute(VIDEO_NAME=video_name),
        output,
    )
    write_scene_list_otio(
        output_path=otio_path,
        scene_list=scenes,
        video_path=str(video_stream.path),
        frame_rate=video_stream.frame_rate,
        name=Template(name).safe_substitute(VIDEO_NAME=video_name),
        audio=audio,
    )



#
#            PySceneDetect: Python-Based Video Scene Detector
#   -------------------------------------------------------------------
#     [  Site:    https://scenedetect.com                           ]
#     [  Docs:    https://scenedetect.com/docs/                     ]
#     [  Github:  https://github.com/Breakthrough/PySceneDetect/    ]
#
# Copyright (C) 2025 Brandon Castellano <http://www.bcastell.com>.
# PySceneDetect is licensed under the BSD 3-Clause License; see the
# included LICENSE file, or visit one of the above pages for details.
#
"""``scenedetect.detector`` Module

This module contains the :class:`SceneDetector` interface, from which all scene detectors in
:mod:`scenedetect.detectors` module are derived from.

The SceneDetector class represents the interface which detection algorithms are expected to provide
in order to be compatible with PySceneDetect.

.. warning::

    This API is still unstable, and changes and design improvements are planned for the v1.0
    release. Instead of just timecodes, detection algorithms will also provide a specific type of
    event (in, out, cut, etc...).
"""

import math
from abc import ABC, abstractmethod
from enum import Enum

import numpy

from scenedetect.common import FrameTimecode, Timecode, TimecodeLike
from scenedetect.stats_manager import StatsManager


class SceneDetector(ABC):
    """Base class to inherit from when implementing a scene detection algorithm.

    This API is not yet stable and subject to change.
    """

    def __init__(self):
        self._stats_manager: StatsManager | None = None

    # Required Methods

    @abstractmethod
    def process_frame(
        self, timecode: FrameTimecode, frame_img: numpy.ndarray
    ) -> list[FrameTimecode]:
        """Process the next frame. `timecode` is assumed to be sequential.

        Arguments:
            timecode: Timecode corresponding to the frame being processed.
            frame_img: Video frame as a 24-bit BGR image.

        Returns:
           List of timecodes where scene cuts have been detected, if any.
        """

    # Optional Methods

    def post_process(self, timecode: FrameTimecode) -> list[FrameTimecode]:
        """Called after there are no more frames to process.

        Arguments:
            timecode: The last position in the video which was read.

        Returns:
           List of timecodes where scene cuts have been detected, if any.
        """
        return []

    @property
    def event_buffer_length(self) -> int:
        """The amount of frames a given event can be buffered for, in time. This must be set to the
        amount of frames a detector might emit an event in the past."""
        return 0

    # Frame Stats/Metrics

    @property
    def stats_manager(self) -> StatsManager | None:
        """Optional :class:`StatsManager <scenedetect.stats_manager.StatsManager>` to use for
        storing frame metrics. When this detector is added to a parent
        :class:`SceneManager <scenedetect.scene_manager.SceneManager>`, then this is set to the
        same :class:`StatsManager <scenedetect.stats_manager.StatsManager>` of the parent - but
        only if it has one itself."""
        return self._stats_manager

    @stats_manager.setter
    def stats_manager(self, value: StatsManager | None):
        self._stats_manager = value

    def get_metrics(self) -> list[str]:
        """Returns a list of all metric names/keys used by this detector.

        Returns:
            List of strings of frame metric key names that will be used by
            the detector when a StatsManager is passed to process_frame.
        """
        return []


class FlashFilter:
    """Filters fast-cuts to enforce minimum scene length."""

    class Mode(Enum):
        """Which mode the filter should use for enforcing minimum scene length."""

        MERGE = 0
        """Merge consecutive cuts shorter than filter length."""
        SUPPRESS = 1
        """Suppress consecutive cuts until the filter length has passed."""

    def __init__(self, mode: Mode, length: TimecodeLike):
        """
        Arguments:
            mode: The mode to use when enforcing `length`.
            length: Minimum scene length. Accepts any :data:`TimecodeLike` value (e.g.
                ``int`` frames, ``float`` seconds, ``str`` such as ``"0.6s"`` /
                ``"00:00:00.600"``, or a :class:`FrameTimecode` / :class:`Timecode`).
        """
        self._mode = mode
        # Frame count (int) and seconds (float) representations of `length`. Exactly one is
        # populated up front; the other is computed on the first frame once the framerate is
        # known. Temporal inputs (float/non-digit str / Timecode / FrameTimecode) populate
        # `_filter_secs`; integer inputs (int/digit str) populate `_filter_length`.
        self._filter_length: int = 0
        self._filter_secs: float | None = None
        if isinstance(length, float):
            self._filter_secs = length
        elif isinstance(length, str) and not length.strip().isdigit():
            self._filter_secs = FrameTimecode(timecode=length, fps=100.0).seconds
        elif isinstance(length, (Timecode, FrameTimecode)):
            self._filter_secs = length.seconds
        else:
            self._filter_length = int(length)
        self._last_above: FrameTimecode | None = None  # Last frame above threshold.
        self._merge_enabled = False  # Used to disable merging until at least one cut was found.
        self._merge_triggered = False  # True when the merge filter is active.
        self._merge_start: FrameTimecode | None = None  # Frame where we started merging.

    @property
    def max_behind(self) -> int:
        if self._mode == FlashFilter.Mode.SUPPRESS:
            return 0
        if self._filter_secs is not None:
            # Estimate using 240fps so the event buffer is large enough for any reasonable input.
            return math.ceil(self._filter_secs * 240.0)
        return self._filter_length

    @property
    def _is_disabled(self) -> bool:
        if self._filter_secs is not None:
            return self._filter_secs <= 0.0
        return self._filter_length <= 0

    def filter(self, timecode: FrameTimecode, above_threshold: bool) -> list[FrameTimecode]:
        if self._is_disabled:
            return [timecode] if above_threshold else []
        if self._last_above is None:
            self._last_above = timecode
        if self._mode == FlashFilter.Mode.MERGE:
            return self._filter_merge(timecode=timecode, above_threshold=above_threshold)
        elif self._mode == FlashFilter.Mode.SUPPRESS:
            return self._filter_suppress(timecode=timecode, above_threshold=above_threshold)
        raise RuntimeError("Unhandled FlashFilter mode.")

    def _filter_suppress(
        self, timecode: FrameTimecode, above_threshold: bool
    ) -> list[FrameTimecode]:
        frame_rate = timecode.frame_rate
        assert frame_rate is not None and frame_rate >= 0
        assert self._last_above is not None
        # Compute the threshold in seconds once from the first frame's framerate. This avoids
        # using an incorrect average fps (e.g. OpenCV on VFR video) on subsequent frames.
        if self._filter_secs is None:
            self._filter_secs = self._filter_length / float(frame_rate)
        min_length_met: bool = (timecode - self._last_above) >= self._filter_secs
        if not (above_threshold and min_length_met):
            return []
        # Both length and threshold requirements were satisfied. Emit the cut, and wait until both
        # requirements are met again.
        self._last_above = timecode
        return [timecode]

    def _filter_merge(self, timecode: FrameTimecode, above_threshold: bool) -> list[FrameTimecode]:
        frame_rate = timecode.frame_rate
        assert frame_rate is not None and frame_rate >= 0
        assert self._last_above is not None
        # Compute the threshold in seconds once from the first frame's framerate.
        if self._filter_secs is None:
            self._filter_secs = self._filter_length / float(frame_rate)
        min_length_met: bool = (timecode - self._last_above) >= self._filter_secs
        # Ensure last frame is always advanced to the most recent one that was above the threshold.
        if above_threshold:
            self._last_above = timecode
        if self._merge_triggered:
            # This frame was under the threshold, see if enough frames passed to disable the filter.
            assert self._merge_start is not None
            if (
                min_length_met
                and not above_threshold
                and (self._last_above - self._merge_start) >= self._filter_secs
            ):
                self._merge_triggered = False
                return [self._last_above]
            # Keep merging until enough frames pass below the threshold.
            return []
        # Wait for next frame above the threshold.
        if not above_threshold:
            return []
        # If we met the minimum length requirement, no merging is necessary.
        if min_length_met:
            # Only allow the merge filter once the first cut is emitted.
            self._merge_enabled = True
            return [timecode]
        # Start merging cuts until the length requirement is met.
        if self._merge_enabled:
            self._merge_triggered = True
            self._merge_start = timecode
        return []



#
#            PySceneDetect: Python-Based Video Scene Detector
#   -------------------------------------------------------------------
#     [  Site:    https://scenedetect.com                           ]
#     [  Docs:    https://scenedetect.com/docs/                     ]
#     [  Github:  https://github.com/Breakthrough/PySceneDetect/    ]
#
# Copyright (C) 2018 Brandon Castellano <http://www.bcastell.com>.
# PySceneDetect is licensed under the BSD 3-Clause License; see the
# included LICENSE file, or visit one of the above pages for details.
#
"""PySceneDetect scenedetect.stats_manager Tests

This file includes unit tests for the scenedetect.stats_manager module (specifically,
the StatsManager object, used to coordinate caching of frame metrics to/from a CSV
file to speed up subsequent calls to detect_scenes on a SceneManager.

These tests rely on the SceneManager, VifdeoStreamCv2, and ContentDetector classes.

These tests also require the testvideo.mp4 (see test_scene_manager.py for download
instructions), however any other valid video file can be used as well by modifying
the fixture test_video_file in conftest.py.

Additionally, these tests will create, write to, and read from files which use names
TEST_STATS_FILE_XXXXXXXXXXXX.csv, where the X's will be replaced with random digits.
These files will be deleted, if possible, after the tests are completed running.
"""

import csv
from pathlib import Path

import pytest

from scenedetect.backends.opencv import VideoStreamCv2
from scenedetect.common import FrameTimecode
from scenedetect.detectors import ContentDetector
from scenedetect.scene_manager import SceneManager
from scenedetect.stats_manager import (
    COLUMN_NAME_FRAME_NUMBER,
    COLUMN_NAME_TIMECODE,
    StatsFileCorrupt,
    StatsManager,
)


def test_metrics():
    """Test StatsManager metric registration/setting/getting with a set of pre-defined
    key-value pairs (metric_dict).
    """
    metric_dict = {"some_metric": 1.2345, "another_metric": 6.7890}
    metric_keys = list(metric_dict.keys())

    stats = StatsManager()
    frame_key = 100
    assert not stats.is_save_required()

    stats.register_metrics(metric_keys)

    assert not stats.is_save_required()

    assert not stats.metrics_exist(frame_key, metric_keys)
    assert stats.get_metrics(frame_key, metric_keys) == [None] * len(metric_keys)

    stats.set_metrics(frame_key, metric_dict)

    assert stats.is_save_required()

    assert stats.metrics_exist(frame_key, metric_keys)
    assert stats.metrics_exist(frame_key, metric_keys[1:])

    assert stats.get_metrics(frame_key, metric_keys) == [
        metric_dict[metric_key] for metric_key in metric_keys
    ]


def test_detector_metrics(test_video_file):
    """Test passing StatsManager to a SceneManager and using it for storing the frame metrics
    from a ContentDetector.
    """
    video = VideoStreamCv2(test_video_file)
    stats_manager = StatsManager()
    scene_manager = SceneManager(stats_manager)
    scene_manager.add_detector(ContentDetector())
    video_fps = video.frame_rate
    duration = FrameTimecode("00:00:05", video_fps)
    scene_manager.auto_downscale = True
    scene_manager.detect_scenes(video=video, duration=duration)
    # Check that metrics were written to the StatsManager.
    assert stats_manager.get_metrics(0, ContentDetector.METRIC_KEYS)


def test_load_empty_stats(tmp_path: Path):
    """Test loading an empty stats file, ensuring it results in no errors."""
    path = tmp_path.joinpath("stats.csv")
    with open(path, "w"):
        pass
    stats_manager = StatsManager()
    stats_manager.load_from_csv(path)


def test_save_no_detect_scenes(tmp_path: Path):
    """Test saving without calling detect_scenes."""
    path = tmp_path.joinpath("stats.csv")
    stats_manager = StatsManager()
    stats_manager.save_to_csv(path)


def test_load_hardcoded_file(tmp_path: Path):
    """Test loading a stats file with some hard-coded data generated by this test case."""

    path = tmp_path.joinpath("stats.csv")
    stats_manager = StatsManager()
    with open(path, "w") as stats_file:
        stats_writer = csv.writer(stats_file, lineterminator="\n")

        some_metric_key = "some_metric"
        some_metric_value = 1.2
        some_frame_key = 100
        base_timecode = FrameTimecode(0, 29.97)
        some_frame_timecode = base_timecode + some_frame_key

        # Write out a valid file.
        stats_writer.writerow([COLUMN_NAME_FRAME_NUMBER, COLUMN_NAME_TIMECODE, some_metric_key])
        stats_writer.writerow(
            [some_frame_key + 1, some_frame_timecode.get_timecode(), str(some_metric_value)]
        )

    stats_manager.load_from_csv(path)

    # Check that we decoded the correct values.
    assert stats_manager.metrics_exist(some_frame_key, [some_metric_key])
    assert stats_manager.get_metrics(some_frame_key, [some_metric_key])[0] == pytest.approx(
        some_metric_value
    )


def test_save_load_from_video(test_video_file, tmp_path: Path):
    """Test generating and saving some frame metrics from TEST_VIDEO_FILE to a file on disk, and
    loading the file back to ensure the loaded frame metrics agree with those that were saved.
    """
    video = VideoStreamCv2(test_video_file)
    stats_manager = StatsManager()
    scene_manager = SceneManager(stats_manager)

    scene_manager.add_detector(ContentDetector())

    video_fps = video.frame_rate
    duration = FrameTimecode("00:00:05", video_fps)

    scene_manager.auto_downscale = True
    scene_manager.detect_scenes(video, duration=duration)

    path = tmp_path.joinpath("stats.csv")
    stats_manager.save_to_csv(csv_file=path)

    metrics = stats_manager.metric_keys

    stats_manager_new = StatsManager()

    stats_manager_new.load_from_csv(path)

    # Compare the first 5 frames. Frame 0 won't have any metrics for this detector.
    for frame in range(1, 5 + 1):
        assert stats_manager.metrics_exist(frame, metrics)
        orig_metrics = stats_manager.get_metrics(frame, metrics)
        new_metrics = stats_manager_new.get_metrics(frame, metrics)
        for i, metric_val in enumerate(orig_metrics):
            assert metric_val == pytest.approx(new_metrics[i])


def test_load_corrupt_stats(tmp_path: Path):
    """Test loading a corrupted stats file created by outputting data in the wrong format."""

    stats_manager = StatsManager()

    path = tmp_path.joinpath("stats.csv")
    with open(path, "w") as stats_file:
        stats_writer = csv.writer(stats_file, lineterminator="\n")

        some_metric_key = "some_metric"
        some_metric_value = str(1.2)
        some_frame_key = 100
        base_timecode = FrameTimecode(0, 29.97)
        some_frame_timecode = base_timecode + some_frame_key

        # Write out some invalid files.

        # File #0: Wrong Header Names [StatsFileCorrupt]
        # Swapped timecode & frame number.
        stats_writer.writerow([COLUMN_NAME_TIMECODE, COLUMN_NAME_FRAME_NUMBER, some_metric_key])
        stats_writer.writerow(
            [some_frame_key, some_frame_timecode.get_timecode(), some_metric_value]
        )

        stats_file.close()

        with pytest.raises(StatsFileCorrupt):
            stats_manager.load_from_csv(path)



#
#            PySceneDetect: Python-Based Video Scene Detector
#   -------------------------------------------------------------------
#     [  Site:    https://scenedetect.com                           ]
#     [  Docs:    https://scenedetect.com/docs/                     ]
#     [  Github:  https://github.com/Breakthrough/PySceneDetect/    ]
#
# Copyright (C) 2026 Brandon Castellano <http://www.bcastell.com>.
# PySceneDetect is licensed under the BSD 3-Clause License; see the
# included LICENSE file, or visit one of the above pages for details.
#
"""Unit tests for the benchmark evaluator. Exercises matching, tolerance, and aggregation logic with
synthetic predictions versus ground-truth lists. Covers TRECVID-SBD style scoring as documented in
``benchmark/README.md``.
"""

from __future__ import annotations

import math
from pathlib import Path

from benchmark.evaluator import (
    EventInterval,
    EventMetrics,
    GroundTruth,
    Prediction,
    _score_fade_transitions,
    _score_hard_cuts,
    evaluate,
    score_video,
)

# --------------------------------------------------------------------- #
# Hard-cut matching (the core of every detector's score)
# --------------------------------------------------------------------- #


def test_hard_exact_match_tolerance_zero():
    m, offsets = _score_hard_cuts(
        predicted_cuts=[10, 20, 30], ground_truth_cuts=[10, 20, 30], tolerance=0
    )
    assert (m.matched, m.false_positives, m.missed) == (3, 0, 0)
    assert offsets == [0, 0, 0]
    assert m.precision == 1.0
    assert m.recall == 1.0
    assert m.f1 == 1.0


def test_hard_tolerance_one_admits_one_frame_offset():
    m, offsets = _score_hard_cuts(predicted_cuts=[11, 19], ground_truth_cuts=[10, 20], tolerance=1)
    assert (m.matched, m.false_positives, m.missed) == (2, 0, 0)
    assert sorted(offsets) == [1, 1]


def test_hard_tolerance_one_rejects_two_frame_offset():
    m, _ = _score_hard_cuts(predicted_cuts=[12], ground_truth_cuts=[10], tolerance=1)
    assert (m.matched, m.false_positives, m.missed) == (0, 1, 1)


def test_hard_greedy_picks_closer_match():
    # Two ground-truth cuts at 10 and 14. Single prediction at 13 is closer
    # to 14. The greedy matcher must claim 14 first; 10 then becomes a miss.
    m, offsets = _score_hard_cuts(predicted_cuts=[13], ground_truth_cuts=[10, 14], tolerance=5)
    assert (m.matched, m.false_positives, m.missed) == (1, 0, 1)
    assert offsets == [1]


def test_hard_equidistant_tie_resolves_deterministically():
    # Prediction at 12 is exactly 2 frames from both ground-truth cuts at 10
    # and 14. Tie-break is by stable sort order (i, j) which prefers the
    # lower ground-truth index.
    m, offsets = _score_hard_cuts(predicted_cuts=[12], ground_truth_cuts=[10, 14], tolerance=5)
    assert (m.matched, m.false_positives, m.missed) == (1, 0, 1)
    assert offsets == [2]


def test_hard_one_to_one_no_double_assignment():
    # Two predictions both within tolerance of a single ground-truth cut.
    # Only one can match; the other is a false positive.
    m, _ = _score_hard_cuts(predicted_cuts=[10, 11], ground_truth_cuts=[10], tolerance=1)
    assert (m.matched, m.false_positives, m.missed) == (1, 1, 0)


def test_hard_empty_inputs():
    m, _ = _score_hard_cuts(predicted_cuts=[], ground_truth_cuts=[], tolerance=0)
    assert (m.matched, m.false_positives, m.missed) == (0, 0, 0)
    # Division-by-zero defenses.
    assert m.precision == 0.0
    assert m.recall == 0.0
    assert m.f1 == 0.0


def test_hard_empty_preds_with_nonempty_gt():
    # No predictions: every ground-truth cut is a miss.
    m, offsets = _score_hard_cuts(predicted_cuts=[], ground_truth_cuts=[10, 20], tolerance=1)
    assert (m.matched, m.false_positives, m.missed) == (0, 0, 2)
    assert offsets == []
    assert m.recall == 0.0


def test_hard_empty_gt_with_nonempty_preds():
    # No ground truth: every prediction is a false positive.
    m, offsets = _score_hard_cuts(predicted_cuts=[10], ground_truth_cuts=[], tolerance=1)
    assert (m.matched, m.false_positives, m.missed) == (0, 1, 0)
    assert offsets == []
    assert m.precision == 0.0


# --------------------------------------------------------------------- #
# Fade transition matching (ClipShots-style typed ground truth)
# --------------------------------------------------------------------- #


def test_fade_pred_inside_interval_is_match():
    m, consumed = _score_fade_transitions(predicted_cuts=[15], intervals=[EventInterval(10, 20)])
    assert (m.matched, m.false_positives, m.missed) == (1, 0, 0)
    assert consumed == {0}


def test_fade_pred_outside_interval_not_consumed():
    m, consumed = _score_fade_transitions(predicted_cuts=[25], intervals=[EventInterval(10, 20)])
    assert (m.matched, m.false_positives, m.missed) == (0, 0, 1)
    assert consumed == set()  # passed through to hard scorer


def test_fade_multiple_preds_in_same_interval():
    # First prediction inside the interval is the match; the second is a
    # false positive. Both are consumed (do not leak to hard matching).
    m, consumed = _score_fade_transitions(
        predicted_cuts=[12, 18], intervals=[EventInterval(10, 20)]
    )
    assert (m.matched, m.false_positives, m.missed) == (1, 1, 0)
    assert consumed == {0, 1}


def test_fade_interval_endpoints_inclusive():
    m, _ = _score_fade_transitions(predicted_cuts=[10, 20], intervals=[EventInterval(10, 20)])
    # Both endpoints hit the same interval, so 1 match + 1 false positive.
    assert (m.matched, m.false_positives, m.missed) == (1, 1, 0)


# --------------------------------------------------------------------- #
# score_video: fade transitions take priority over hard cuts
# --------------------------------------------------------------------- #


def test_score_video_fade_consumes_pred_before_hard():
    # Prediction at 15 falls inside the fade interval [10, 20]. Even
    # though the hard ground-truth cut at 16 is within tolerance, the
    # fade scorer claims the prediction first and the hard scorer
    # never sees it.
    ground_truth = GroundTruth(hard_cuts=[16], fades=[EventInterval(10, 20)])
    v = score_video([15], ground_truth, tolerance=1, elapsed=0.0)
    assert v.fades.matched == 1
    assert v.hard_cuts.matched == 0  # hard match was preempted by the fade
    assert v.hard_cuts.missed == 1  # hard ground-truth cut at 16 is now a miss


def test_score_video_pred_outside_fade_falls_to_hard():
    ground_truth = GroundTruth(hard_cuts=[30], fades=[EventInterval(10, 20)])
    v = score_video([30], ground_truth, tolerance=0, elapsed=0.0)
    assert v.fades.matched == 0
    assert v.fades.missed == 1  # fade still missed
    assert v.hard_cuts.matched == 1


# --------------------------------------------------------------------- #
# Mean absolute offset (localization error on hard-cut matches only)
# --------------------------------------------------------------------- #


def test_mean_abs_offset_only_hard_matches_tolerance_zero():
    ground_truth = GroundTruth(
        hard_cuts=[100, 200, 300],
        fades=[EventInterval(50, 60)],
    )
    # Predictions: fade hit at 55 (excluded from offset), hard match at 100
    # (offset 0). 201 and 302 are outside tolerance 0.
    v = score_video([55, 100, 201, 302], ground_truth, 0, 0.0)
    assert v.hard_cuts.matched == 1
    assert v.mean_abs_offset == 0.0


def test_mean_abs_offset_only_hard_matches_tolerance_one():
    ground_truth = GroundTruth(
        hard_cuts=[100, 200, 300],
        fades=[EventInterval(50, 60)],
    )
    # Same setup at tolerance 1: 100 (offset 0) and 201 (offset 1) match;
    # 302 is out of tolerance. Mean offset is (0 + 1) / 2 = 0.5.
    v = score_video([55, 100, 201, 302], ground_truth, 1, 0.0)
    assert v.hard_cuts.matched == 2
    assert v.mean_abs_offset == 0.5


def test_mean_abs_offset_nan_when_no_matches():
    ground_truth = GroundTruth(hard_cuts=[1000])
    v = score_video([5], ground_truth, 0, 0.0)
    assert math.isnan(v.mean_abs_offset)


def test_benchmark_result_mean_abs_offset_nan_when_no_matches_across_videos():
    # Two videos, both producing zero hard-cut matches. The aggregate offset
    # has zero sum and zero count, so nan must propagate at the
    # BenchmarkResult level, not just per-video.
    predictions = {
        Path("a.mp4"): Prediction(
            predicted_cuts=[5],
            ground_truth=GroundTruth(hard_cuts=[1000]),
            elapsed=1.0,
        ),
        Path("b.mp4"): Prediction(
            predicted_cuts=[7],
            ground_truth=GroundTruth(hard_cuts=[2000]),
            elapsed=1.0,
        ),
    }
    result = evaluate(predictions, tolerance=0)
    assert math.isnan(result.mean_abs_offset_hard_cuts)


# --------------------------------------------------------------------- #
# Aggregate result: sum-of-counts across videos
# --------------------------------------------------------------------- #


def test_benchmark_result_aggregate_matches_sum_of_counts():
    predictions = {
        Path("vid_a.mp4"): Prediction(
            predicted_cuts=[10, 20],
            ground_truth=GroundTruth(hard_cuts=[10, 21]),
            elapsed=1.0,
        ),
        Path("vid_b.mp4"): Prediction(
            predicted_cuts=[50, 99],
            ground_truth=GroundTruth(hard_cuts=[50, 100]),
            elapsed=3.0,
        ),
    }
    # Tolerance 0: only 10 (vid_a) and 50 (vid_b) match exactly.
    # Aggregate: 2 matched, 2 false positives, 2 missed.
    result_t0 = evaluate(predictions, tolerance=0)
    assert result_t0.hard_cuts.matched == 2
    assert result_t0.hard_cuts.false_positives == 2
    assert result_t0.hard_cuts.missed == 2
    # Tolerance 1: both predictions in each video match -> 4 matched, 0 fp, 0 missed.
    result_t1 = evaluate(predictions, tolerance=1)
    assert result_t1.hard_cuts.matched == 4
    assert result_t1.hard_cuts.false_positives == 0
    assert result_t1.hard_cuts.missed == 0
    # Elapsed: total and mean (independent of tolerance).
    assert result_t0.elapsed_total == 4.0
    assert result_t0.elapsed_mean == 2.0


def test_benchmark_result_by_category_buckets_videos():
    predictions = {
        Path("a.mp4"): Prediction(
            predicted_cuts=[10],
            ground_truth=GroundTruth(hard_cuts=[10], category="news"),
            elapsed=1.0,
        ),
        Path("b.mp4"): Prediction(
            predicted_cuts=[20],
            ground_truth=GroundTruth(hard_cuts=[20], category="sports"),
            elapsed=1.0,
        ),
        Path("c.mp4"): Prediction(
            predicted_cuts=[30],
            ground_truth=GroundTruth(hard_cuts=[30], category="news"),
            elapsed=1.0,
        ),
    }
    result = evaluate(predictions, tolerance=0)
    by_category = result.by_category()
    assert set(by_category) == {"news", "sports"}
    assert len(by_category["news"].per_video) == 2
    assert len(by_category["sports"].per_video) == 1


def test_benchmark_result_by_category_buckets_untagged_videos_as_unknown():
    # Datasets without category tags (BBC, AutoShot) leave category=None on every
    # video. by_category must bucket those under the literal key "unknown".
    predictions = {
        Path("a.mp4"): Prediction(
            predicted_cuts=[10],
            ground_truth=GroundTruth(hard_cuts=[10]),  # category defaults to None
            elapsed=1.0,
        ),
        Path("b.mp4"): Prediction(
            predicted_cuts=[20],
            ground_truth=GroundTruth(hard_cuts=[20]),
            elapsed=1.0,
        ),
    }
    result = evaluate(predictions, tolerance=0)
    by_category = result.by_category()
    assert set(by_category) == {"unknown"}
    assert len(by_category["unknown"].per_video) == 2


# --------------------------------------------------------------------- #
# EventMetrics arithmetic
# --------------------------------------------------------------------- #


def test_event_metrics_addition():
    a = EventMetrics(matched=3, false_positives=1, missed=2)
    b = EventMetrics(matched=5, false_positives=2, missed=1)
    c = a + b
    assert (c.matched, c.false_positives, c.missed) == (8, 3, 3)


def test_event_metrics_to_dict_round_trip():
    m = EventMetrics(matched=3, false_positives=1, missed=1)
    d = m.to_dict()
    assert d["matched"] == 3
    assert d["false_positives"] == 1
    assert d["missed"] == 1
    assert d["precision"] == 75.0  # 3 / 4
    assert d["recall"] == 75.0  # 3 / 4
    assert d["f1"] == 75.0



#
#            PySceneDetect: Python-Based Video Scene Detector
#   -------------------------------------------------------------------
#     [  Site:    https://scenedetect.com                           ]
#     [  Docs:    https://scenedetect.com/docs/                     ]
#     [  Github:  https://github.com/Breakthrough/PySceneDetect/    ]
#
# Copyright (C) 2025 Brandon Castellano <http://www.bcastell.com>.
# PySceneDetect is licensed under the BSD 3-Clause License; see the
# included LICENSE file, or visit one of the above pages for details.
#

"""The ``scenedetect.output`` module contains functions which can be used to generate output
based on the output of scene detection. This includes saving images for each scene, exporting to
CSV/HTML, or splitting the input video into individual shots.
"""

import csv
import json
import logging
import math
import typing as ty
from fractions import Fraction
from pathlib import Path
from xml.dom import minidom
from xml.etree import ElementTree

from scenedetect._thirdparty.simpletable import (
    HTMLPage,
    SimpleTable,
    SimpleTableCell,
    SimpleTableImage,
    SimpleTableRow,
)
from scenedetect.common import (
    CutList,
    FrameTimecode,
    SceneList,
)

# Commonly used classes/functions exported under the `scenedetect.output` namespace for brevity.
from scenedetect.output.image import save_images as save_images
from scenedetect.output.video import (
    PathFormatter as PathFormatter,
)
from scenedetect.output.video import (
    SceneMetadata as SceneMetadata,
)
from scenedetect.output.video import (
    VideoMetadata as VideoMetadata,
)
from scenedetect.output.video import (
    default_formatter as default_formatter,
)
from scenedetect.output.video import (
    is_ffmpeg_available as is_ffmpeg_available,
)
from scenedetect.output.video import (
    is_mkvmerge_available as is_mkvmerge_available,
)
from scenedetect.output.video import (
    split_video_ffmpeg as split_video_ffmpeg,
)
from scenedetect.output.video import (
    split_video_mkvmerge as split_video_mkvmerge,
)

logger = logging.getLogger("pyscenedetect")


def write_scene_list(
    output_csv_file: ty.TextIO,
    scene_list: SceneList,
    include_cut_list: bool = True,
    cut_list: CutList | None = None,
    col_separator: str = ",",
    row_separator: str = "\n",
):
    """Writes the given list of scenes to an output file handle in CSV format.

    Arguments:
        output_csv_file: Handle to open file in write mode.
        scene_list: List of pairs of FrameTimecodes denoting each scene's start/end FrameTimecode.
        include_cut_list: Bool indicating if the first row should include the timecodes where
            each scene starts. Should be set to False if RFC 4180 compliant CSV output is required.
        cut_list: Optional list of FrameTimecode objects denoting the cut list (i.e. the frames
            in the video that need to be split to generate individual scenes). If not specified,
            the cut list is generated using the start times of each scene following the first one.
        col_separator: Delimiter to use between values. Must be single character.
        row_separator: Line terminator to use between rows.

    Raises:
        TypeError: "delimiter" must be a 1-character string
    """
    csv_writer = csv.writer(output_csv_file, delimiter=col_separator, lineterminator=row_separator)
    # If required, output the cutting list as the first row (i.e. before the header row).
    if include_cut_list:
        csv_writer.writerow(
            ["Timecode List:", *cut_list]
            if cut_list
            else [start.get_timecode() for start, _ in scene_list[1:]]
        )
    csv_writer.writerow(
        [
            "Scene Number",
            "Start Frame",
            "Start Timecode",
            "Start Time (seconds)",
            "End Frame",
            "End Timecode",
            "End Time (seconds)",
            "Length (frames)",
            "Length (timecode)",
            "Length (seconds)",
        ]
    )
    for i, (start, end) in enumerate(scene_list):
        duration = end - start
        csv_writer.writerow(
            [
                f"{i + 1:d}",
                f"{start.frame_num + 1:d}",
                start.get_timecode(),
                f"{start.seconds:.3f}",
                f"{end.frame_num:d}",
                end.get_timecode(),
                f"{end.seconds:.3f}",
                f"{duration.frame_num:d}",
                duration.get_timecode(),
                f"{duration.seconds:.3f}",
            ]
        )


def write_scene_list_html(
    output_html_filename: str,
    scene_list: SceneList,
    cut_list: CutList | None = None,
    css: str | None = None,
    css_class: str = "mytable",
    image_filenames: dict[int, list[str]] | None = None,
    image_width: int | None = None,
    image_height: int | None = None,
):
    """Writes the given list of scenes to an output file handle in html format.

    Arguments:
        output_html_filename: filename of output html file
        scene_list: List of pairs of FrameTimecodes denoting each scene's start/end FrameTimecode.
        cut_list: Optional list of FrameTimecode objects denoting the cut list (i.e. the frames
            in the video that need to be split to generate individual scenes). If not passed,
            the start times of each scene (besides the 0th scene) is used instead.
        css: String containing all the css information for the resulting html page.
        css_class: String containing the named css class
        image_filenames: dict where key i contains a list with n elements (filenames of
            the n saved images from that scene)
        image_width: Optional desired width of images in table in pixels
        image_height: Optional desired height of images in table in pixels
    """
    logger.info("Exporting scenes to html:\n %s:", output_html_filename)
    if not css:
        css = """
        table.mytable {
            font-family: times;
            font-size:12px;
            color:#000000;
            border-width: 1px;
            border-color: #eeeeee;
            border-collapse: collapse;
            background-color: #ffffff;
            width=100%;
            max-width:550px;
            table-layout:fixed;
        }
        table.mytable th {
            border-width: 1px;
            padding: 8px;
            border-style: solid;
            border-color: #eeeeee;
            background-color: #e6eed6;
            color:#000000;
        }
        table.mytable td {
            border-width: 1px;
            padding: 8px;
            border-style: solid;
            border-color: #eeeeee;
        }
        #code {
            display:inline;
            font-family: courier;
            color: #3d9400;
        }
        #string {
            display:inline;
            font-weight: bold;
        }
        """

    # Output Timecode list
    timecode_table = SimpleTable(
        [
            ["Timecode List:"]
            + (cut_list if cut_list else [start.get_timecode() for start, _ in scene_list[1:]])
        ],
        css_class=css_class,
    )

    # Output list of scenes
    header_row = [
        "Scene Number",
        "Start Frame",
        "Start Timecode",
        "Start Time (seconds)",
        "End Frame",
        "End Timecode",
        "End Time (seconds)",
        "Length (frames)",
        "Length (timecode)",
        "Length (seconds)",
    ]
    for i, (start, end) in enumerate(scene_list):
        duration = end - start

        row = SimpleTableRow(
            [
                f"{i + 1:d}",
                f"{start.frame_num + 1:d}",
                start.get_timecode(),
                f"{start.seconds:.3f}",
                f"{end.frame_num:d}",
                end.get_timecode(),
                f"{end.seconds:.3f}",
                f"{duration.frame_num:d}",
                duration.get_timecode(),
                f"{duration.seconds:.3f}",
            ]
        )

        if image_filenames:
            for image in image_filenames[i]:
                row.add_cell(
                    SimpleTableCell(SimpleTableImage(image, width=image_width, height=image_height))
                )

        if i == 0:
            scene_table = SimpleTable(rows=[row], header_row=header_row, css_class=css_class)
        else:
            scene_table.add_row(row=row)

    # Write html file
    page = HTMLPage()
    page.add_table(timecode_table)
    page.add_table(scene_table)
    page.css = css
    page.save(output_html_filename)


def _edl_timecode(timecode: FrameTimecode) -> str:
    """Format `timecode` as ``HH:MM:SS:FF`` for a CMX 3600 EDL entry."""
    total_seconds = timecode.seconds
    frame_rate = timecode.frame_rate
    assert frame_rate is not None
    hours = int(total_seconds // 3600)
    minutes = int((total_seconds % 3600) // 60)
    seconds = int(total_seconds % 60)
    frames_part = int((total_seconds * frame_rate) % frame_rate)
    return f"{hours:02d}:{minutes:02d}:{seconds:02d}:{frames_part:02d}"


def _parse_edl_start_timecode(value: str, frame_rate: Fraction | float) -> int:
    """Parse a SMPTE ``HH:MM:SS:FF`` (or 8-digit ``HHMMSSFF``) start timecode into a frame count."""
    stripped = value.strip()
    if ":" in stripped:
        parts = stripped.split(":")
    elif stripped.isdigit() and len(stripped) == 8:
        parts = [stripped[0:2], stripped[2:4], stripped[4:6], stripped[6:8]]
    else:
        raise ValueError(
            f"Invalid start timecode {value!r}: expected HH:MM:SS:FF or 8 digits (HHMMSSFF)."
        )
    if len(parts) != 4 or not all(p.isdigit() for p in parts):
        raise ValueError(
            f"Invalid start timecode {value!r}: expected HH:MM:SS:FF or 8 digits (HHMMSSFF)."
        )
    hours, minutes, seconds, frames = (int(p) for p in parts)
    max_frames = math.ceil(float(frame_rate))
    if minutes >= 60 or seconds >= 60 or frames >= max_frames:
        raise ValueError(
            f"Invalid start timecode {value!r}: MM<60, SS<60, FF<{max_frames} required."
        )
    return round((hours * 3600 + minutes * 60 + seconds) * float(frame_rate)) + frames


def write_scene_list_edl(
    output_path: str | Path,
    scene_list: SceneList,
    title: str = "PySceneDetect",
    reel: str = "AX",
    start_timecode: str | None = None,
):
    """Writes the given list of scenes to `output_path` in CMX 3600 EDL format.

    Arguments:
        output_path: Path to write the EDL file to. Parent directories must exist.
        scene_list: List of scenes as pairs of FrameTimecodes denoting each scene's start/end.
        title: Title header written as ``TITLE:`` in the EDL.
        reel: Reel name used for each event. Typically 2-8 uppercase characters.
        start_timecode: Optional SMPTE timecode (``HH:MM:SS:FF`` or 8-digit ``HHMMSSFF``) added to
            every event so the EDL aligns with the source media's on-screen timecode. Applied to
            both source and record columns.
    """
    output_path = Path(output_path)
    offset_frames = 0
    if start_timecode is not None and start_timecode.strip() and scene_list:
        frame_rate = scene_list[0][0].frame_rate
        assert frame_rate is not None
        offset_frames = _parse_edl_start_timecode(start_timecode, frame_rate)
    lines = [f"TITLE: {title}", "FCM: NON-DROP FRAME", ""]
    for i, (start, end) in enumerate(scene_list):
        in_tc = _edl_timecode(start + offset_frames)
        out_tc = _edl_timecode(end + offset_frames)
        lines.append(f"{(i + 1):03d}  {reel} V     C        {in_tc} {out_tc} {in_tc} {out_tc}")
    logger.info("Writing scenes in EDL format to %s", output_path)
    with open(output_path, "w") as f:
        # `scenedetect` is imported lazily to avoid a circular import at module load.
        import scenedetect

        f.write(f"* CREATED WITH PYSCENEDETECT {scenedetect.__version__}\n")
        f.write("\n".join(lines))
        f.write("\n")


def _rational_seconds(value: Fraction) -> str:
    """Format a `Fraction` as an FCPXML rational time string.

    FCPXML expresses time as ``<num>/<denom>s`` (or ``<int>s`` for whole seconds). See
    https://developer.apple.com/documentation/professional-video-applications/fcpxml-reference
    """
    if value.denominator == 1:
        return f"{value.numerator}s"
    return f"{value.numerator}/{value.denominator}s"


def _frame_timecode_seconds(tc: FrameTimecode) -> Fraction:
    """Exact seconds for `tc` as a `Fraction`, derived from PTS * time base."""
    return Fraction(tc.pts) * tc.time_base


def write_scene_list_fcpx(
    output_path: str | Path,
    scene_list: SceneList,
    video_path: str | Path,
    frame_rate: Fraction,
    frame_size: tuple[int, int],
    video_name: str | None = None,
):
    """Writes the given list of scenes to `output_path` in Final Cut Pro X XML format (FCPXML 1.9).

    The output follows Apple's FCPXML schema with rational-second time values and a custom
    ``<format>`` derived from the source video's frame rate and resolution. See
    https://developer.apple.com/documentation/professional-video-applications/fcpxml-reference

    Arguments:
        output_path: Path to write the FCPXML file to. Parent directories must exist.
        scene_list: List of scenes as pairs of FrameTimecodes. Must not be empty.
        video_path: Path to the source video file; written into the output as a ``file://`` URI.
        frame_rate: Source frame rate as a rational `Fraction` (e.g. ``Fraction(24000, 1001)``).
        frame_size: Source resolution as a ``(width, height)`` tuple in pixels.
        video_name: Display name used for the asset, project, and event. Defaults to the stem
            of `video_path`.
    """
    assert scene_list
    output_path = Path(output_path)
    video_path = Path(video_path)
    if video_name is None:
        video_name = video_path.stem

    ASSET_ID = "r2"
    FORMAT_ID = "r1"

    width, height = frame_size
    frame_duration = _rational_seconds(Fraction(frame_rate.denominator, frame_rate.numerator))
    src_uri = video_path.absolute().as_uri()
    total_duration = _rational_seconds(
        _frame_timecode_seconds(scene_list[-1][1] - scene_list[0][0])
    )

    root = ElementTree.Element("fcpxml", version="1.9")
    resources = ElementTree.SubElement(root, "resources")
    # `name` is cosmetic: Apple publishes no authoritative FFVideoFormat* list, and editors key
    # off frameDuration/width/height. We emit a generated name for display only.
    format_name = f"FFVideoFormat{height}p{round(float(frame_rate) * 100):04d}"
    ElementTree.SubElement(
        resources,
        "format",
        id=FORMAT_ID,
        name=format_name,
        frameDuration=frame_duration,
        width=str(width),
        height=str(height),
    )
    asset = ElementTree.SubElement(
        resources,
        "asset",
        id=ASSET_ID,
        name=video_name,
        start="0s",
        duration=total_duration,
        hasVideo="1",
        format=FORMAT_ID,
    )
    ElementTree.SubElement(asset, "media-rep", kind="original-media", src=src_uri)

    library = ElementTree.SubElement(root, "library")
    event = ElementTree.SubElement(library, "event", name=video_name)
    project = ElementTree.SubElement(event, "project", name=video_name)
    sequence = ElementTree.SubElement(
        project,
        "sequence",
        format=FORMAT_ID,
        duration=total_duration,
        tcStart="0s",
        tcFormat="NDF",
    )
    spine = ElementTree.SubElement(sequence, "spine")

    for i, (start, end) in enumerate(scene_list):
        scene_start = _rational_seconds(_frame_timecode_seconds(start))
        scene_duration = _rational_seconds(_frame_timecode_seconds(end - start))
        ElementTree.SubElement(
            spine,
            "asset-clip",
            name=f"Shot {i + 1}",
            ref=ASSET_ID,
            offset=scene_start,
            start=scene_start,
            duration=scene_duration,
        )

    pretty_xml = minidom.parseString(ElementTree.tostring(root, encoding="unicode")).toprettyxml(
        indent="  "
    )
    logger.info("Writing scenes in FCPX format to %s", output_path)
    with open(output_path, "w") as f:
        f.write(pretty_xml)


def write_scene_list_fcp7(
    output_path: str | Path,
    scene_list: SceneList,
    video_path: str | Path,
    frame_rate: Fraction,
    frame_size: tuple[int, int],
    video_name: str | None = None,
    source_duration: FrameTimecode | None = None,
):
    """Writes the given list of scenes to `output_path` in Final Cut Pro 7 XML (xmeml) format.

    See the xmeml element reference at
    https://developer.apple.com/library/archive/documentation/AppleApplications/Reference/FinalCutPro_XML/.
    ``pathurl`` is written as a valid ``file://`` URI per the xmeml spec.

    Arguments:
        output_path: Path to write the xmeml file to. Parent directories must exist.
        scene_list: List of scenes as pairs of FrameTimecodes. Must not be empty.
        video_path: Path to the source video file; written into the output as a ``file://`` URI.
        frame_rate: Source frame rate as a rational `Fraction`.
        frame_size: Source resolution as a ``(width, height)`` tuple in pixels.
        video_name: Display name used for project and sequence. Defaults to the stem of
            `video_path`.
        source_duration: Total duration of the source media. Required on ``<file>`` so NLEs
            (DaVinci Resolve, Premiere) can seek into the source - without it the clip plays
            frozen. If None, falls back to the last scene's end time.
    """
    assert scene_list
    output_path = Path(output_path)
    video_path = Path(video_path)
    if video_name is None:
        video_name = video_path.stem

    root = ElementTree.Element("xmeml", version="5")
    project = ElementTree.SubElement(root, "project")
    ElementTree.SubElement(project, "name").text = video_name
    sequence = ElementTree.SubElement(project, "sequence")
    ElementTree.SubElement(sequence, "name").text = video_name

    fps = float(frame_rate)
    ntsc = "True" if frame_rate.denominator != 1 else "False"
    duration = scene_list[-1][1] - scene_list[0][0]
    ElementTree.SubElement(sequence, "duration").text = str(round(duration.seconds * fps))

    rate = ElementTree.SubElement(sequence, "rate")
    ElementTree.SubElement(rate, "timebase").text = str(round(fps))
    ElementTree.SubElement(rate, "ntsc").text = ntsc

    timecode = ElementTree.SubElement(sequence, "timecode")
    tc_rate = ElementTree.SubElement(timecode, "rate")
    ElementTree.SubElement(tc_rate, "timebase").text = str(round(fps))
    ElementTree.SubElement(tc_rate, "ntsc").text = ntsc
    ElementTree.SubElement(timecode, "frame").text = "0"
    ElementTree.SubElement(timecode, "displayformat").text = "NDF"

    width, height = frame_size
    media = ElementTree.SubElement(sequence, "media")
    video = ElementTree.SubElement(media, "video")
    format = ElementTree.SubElement(video, "format")
    sample_chars = ElementTree.SubElement(format, "samplecharacteristics")
    ElementTree.SubElement(sample_chars, "width").text = str(width)
    ElementTree.SubElement(sample_chars, "height").text = str(height)
    track = ElementTree.SubElement(video, "track")

    path_uri = video_path.absolute().as_uri()
    source_duration_frames = str(
        round(
            (source_duration.seconds if source_duration is not None else scene_list[-1][1].seconds)
            * fps
        )
    )
    FILE_ID = "file1"

    for i, (start, end) in enumerate(scene_list):
        clip = ElementTree.SubElement(track, "clipitem")
        ElementTree.SubElement(clip, "name").text = f"Shot {i + 1}"
        ElementTree.SubElement(clip, "enabled").text = "TRUE"
        ElementTree.SubElement(clip, "duration").text = source_duration_frames
        clip_rate = ElementTree.SubElement(clip, "rate")
        ElementTree.SubElement(clip_rate, "timebase").text = str(round(fps))
        ElementTree.SubElement(clip_rate, "ntsc").text = ntsc
        # Frame numbers relative to the declared <timebase> fps, computed from PTS seconds.
        ElementTree.SubElement(clip, "start").text = str(round(start.seconds * fps))
        ElementTree.SubElement(clip, "end").text = str(round(end.seconds * fps))
        ElementTree.SubElement(clip, "in").text = str(round(start.seconds * fps))
        ElementTree.SubElement(clip, "out").text = str(round(end.seconds * fps))

        # xmeml allows a single full `<file>` declaration reused via `<file id="...">` on
        # subsequent clipitems. Emit full details on the first, then self-close on the rest.
        if i == 0:
            file_ref = ElementTree.SubElement(clip, "file", id=FILE_ID)
            ElementTree.SubElement(file_ref, "name").text = video_name
            ElementTree.SubElement(file_ref, "pathurl").text = path_uri
            ElementTree.SubElement(file_ref, "duration").text = source_duration_frames
            file_rate = ElementTree.SubElement(file_ref, "rate")
            ElementTree.SubElement(file_rate, "timebase").text = str(round(fps))
            ElementTree.SubElement(file_rate, "ntsc").text = ntsc
            media_ref = ElementTree.SubElement(file_ref, "media")
            video_ref = ElementTree.SubElement(media_ref, "video")
            clip_chars = ElementTree.SubElement(video_ref, "samplecharacteristics")
            ElementTree.SubElement(clip_chars, "width").text = str(width)
            ElementTree.SubElement(clip_chars, "height").text = str(height)
        else:
            ElementTree.SubElement(clip, "file", id=FILE_ID)

        link = ElementTree.SubElement(clip, "link")
        ElementTree.SubElement(link, "linkclipref").text = FILE_ID
        ElementTree.SubElement(link, "mediatype").text = "video"

    pretty_xml = minidom.parseString(ElementTree.tostring(root, encoding="unicode")).toprettyxml(
        indent="  "
    )
    logger.info("Writing scenes in FCP format to %s", output_path)
    with open(output_path, "w") as f:
        f.write(pretty_xml)


# TODO: We have to export framerate as a float for OTIO's current format. When OTIO supports
# fractional timecodes, we should export the framerate as a rational number instead.
# https://github.com/AcademySoftwareFoundation/OpenTimelineIO/issues/190
def write_scene_list_otio(
    output_path: str | Path,
    scene_list: SceneList,
    video_path: str | Path,
    frame_rate: Fraction,
    name: str | None = None,
    audio: bool = True,
):
    """Writes the given list of scenes to `output_path` as an OTIO Timeline.1 JSON document.

    OTIO (OpenTimelineIO) timelines can be imported by many video editors.

    Arguments:
        output_path: Path to write the OTIO file to. Parent directories must exist.
        scene_list: List of scenes as pairs of FrameTimecodes.
        video_path: Path to the source video file; written into the output as an absolute path.
        frame_rate: Source frame rate as a rational `Fraction`. Exported as a float, as the
            current OTIO format does not support rational timings.
        name: Timeline name. Defaults to the stem of `video_path`.
        audio: If True (default), include an audio track alongside the video track.
    """
    output_path = Path(output_path)
    video_path = Path(video_path)
    if name is None:
        name = video_path.stem

    video_base_name = video_path.name
    video_abs_path = str(video_path.absolute())
    fps = float(frame_rate)

    # List of track mapping to resource type.
    # TODO(https://scenedetect.com/issues/497): Allow OTIO export without an audio track.
    track_list = {"Video 1": "Video"}
    if audio:
        track_list["Audio 1"] = "Audio"

    otio = {
        "OTIO_SCHEMA": "Timeline.1",
        "name": name,
        "global_start_time": {
            "OTIO_SCHEMA": "RationalTime.1",
            "rate": fps,
            "value": 0.0,
        },
        "tracks": {
            "OTIO_SCHEMA": "Stack.1",
            "enabled": True,
            "children": [
                {
                    "OTIO_SCHEMA": "Track.1",
                    "name": track_name,
                    "enabled": True,
                    "children": [
                        {
                            "OTIO_SCHEMA": "Clip.2",
                            "name": video_base_name,
                            "source_range": {
                                "OTIO_SCHEMA": "TimeRange.1",
                                "duration": {
                                    "OTIO_SCHEMA": "RationalTime.1",
                                    "rate": fps,
                                    "value": round((end - start).seconds * fps, 6),
                                },
                                "start_time": {
                                    "OTIO_SCHEMA": "RationalTime.1",
                                    "rate": fps,
                                    "value": round(start.seconds * fps, 6),
                                },
                            },
                            "enabled": True,
                            "media_references": {
                                "DEFAULT_MEDIA": {
                                    "OTIO_SCHEMA": "ExternalReference.1",
                                    "name": video_base_name,
                                    "available_range": {
                                        "OTIO_SCHEMA": "TimeRange.1",
                                        "duration": {
                                            "OTIO_SCHEMA": "RationalTime.1",
                                            "rate": fps,
                                            "value": 1980.0,
                                        },
                                        "start_time": {
                                            "OTIO_SCHEMA": "RationalTime.1",
                                            "rate": fps,
                                            "value": 0.0,
                                        },
                                    },
                                    "available_image_bounds": None,
                                    "target_url": video_abs_path,
                                }
                            },
                            "active_media_reference_key": "DEFAULT_MEDIA",
                        }
                        for (start, end) in scene_list
                    ],
                    "kind": track_type,
                }
                for (track_name, track_type) in track_list.items()
            ],
        },
    }

    logger.info("Writing scenes in OTIO format to %s", output_path)
    with open(output_path, "w") as f:
        json.dump(otio, f, indent=4)
        f.write("\n")



#
#         PySceneDetect: Python-Based Video Scene Detector
#   ---------------------------------------------------------------
#     [  Site:   http://www.scenedetect.scenedetect.com/         ]
#     [  Docs:   http://manual.scenedetect.scenedetect.com/      ]
#     [  Github: https://github.com/Breakthrough/PySceneDetect/  ]
#
# Copyright (C) 2024 Brandon Castellano <http://www.bcastell.com>.
# PySceneDetect is licensed under the BSD 3-Clause License; see the
# included LICENSE file, or visit one of the above pages for details.
#
""":py:class:`HistogramDetector` compares the difference in the YUV histograms of subsequent
frames. If the difference exceeds a given threshold, a cut is detected.

This detector is available from the command-line as the `detect-hist` command.
"""

import typing as ty

import cv2
import numpy

from scenedetect.common import FrameTimecode, TimecodeLike
from scenedetect.detector import SceneDetector


class HistogramDetector(SceneDetector):
    """Compares the difference in the Y channel of YUV histograms for adjacent frames. When the
    difference exceeds a given threshold, a cut is detected."""

    METRIC_KEYS: ty.ClassVar[list[str]] = ["hist_diff"]

    def __init__(
        self,
        threshold: float = 0.05,
        bins: int = 256,
        min_scene_len: TimecodeLike = 15,
    ):
        """
        Arguments:
            threshold: maximum relative difference between 0.0 and 1.0 that the histograms can
                differ. Histograms are calculated on the Y channel after converting the frame to
                YUV, and normalized based on the number of bins. Higher differences imply greater
                change in content, so larger threshold values are less sensitive to cuts.
            bins: Number of bins to use for the histogram.
            min_scene_len: Once a cut is detected, this much time must pass before a new one can
                be added to the scene list. Accepts any :data:`TimecodeLike` value.
        """
        super().__init__()
        # Internally, threshold represents the correlation between two histograms and has values
        # between -1.0 and 1.0.
        self._threshold = max(0.0, min(1.0, 1.0 - threshold))
        self._bins = bins
        self._min_scene_len = min_scene_len
        self._last_hist = None
        self._last_cut = None
        self._metric_key = f"hist_diff [bins={self._bins}]"

    def process_frame(
        self, timecode: FrameTimecode, frame_img: numpy.ndarray
    ) -> list[FrameTimecode]:
        """Computes the histogram of the luma channel of the frame image and compares it with the
        histogram of the luma channel of the previous frame. If the difference between the
        histograms exceeds the threshold, a scene cut is detected.
        Histogram difference is computed using the correlation metric.

        Arguments:
            timecode: Timecode of the frame that is being passed.
            frame_img: Decoded frame image (numpy.ndarray) to perform scene
                detection on.

        Returns:
            List of timecodes where scene cuts have been detected. There may be 0
            or more timecodes in the list, and not necessarily the same as `timecode`.
        """
        cut_list = []

        np_data_type = frame_img.dtype

        if np_data_type != numpy.uint8:
            raise ValueError("Image must be 8-bit rgb for HistogramDetector")

        if frame_img.shape[2] != 3:
            raise ValueError("Image must have three color channels for HistogramDetector")

        # Initialize last scene cut point at the beginning of the frames of interest.
        if not self._last_cut:
            self._last_cut = timecode

        hist = self.calculate_histogram(frame_img, bins=self._bins)

        # We can only start detecting once we have a frame to compare with.
        if self._last_hist is not None:
            # TODO: We can have EMA of histograms to make it more robust
            # ema_hist = alpha * hist + (1 - alpha) * ema_hist

            # Compute histogram difference between frames
            hist_diff = cv2.compareHist(self._last_hist, hist, cv2.HISTCMP_CORREL)

            # Check if a new scene should be triggered
            # Set a correlation threshold to determine scene changes.
            # The threshold value should be between -1 (perfect negative correlation, not
            # applicable here) and +1 (perfect positive correlation, identical histograms).
            # Values close to 1 indicate very similar frames, while lower values suggest changes.
            # Example: If `_threshold` is set to 0.8, it implies that only changes resulting in a
            # correlation less than 0.8 between histograms will be considered significant enough to
            # denote a scene change.
            if hist_diff <= self._threshold and (
                (timecode - self._last_cut) >= self._min_scene_len
            ):
                cut_list.append(timecode)
                self._last_cut = timecode

            # Save stats to a StatsManager if it is being used
            if self.stats_manager is not None:
                self.stats_manager.set_metrics(timecode, {self._metric_key: hist_diff})

        self._last_hist = hist

        return cut_list

    @staticmethod
    def calculate_histogram(
        frame_img: numpy.ndarray, bins: int = 256, normalize: bool = True
    ) -> numpy.ndarray:
        """
        Calculates and optionally normalizes the histogram of the luma (Y) channel of an image
        converted from BGR to YUV color space.

        This function extracts the Y channel from the given BGR image, computes its histogram with
        the specified number of bins, and optionally normalizes this histogram to have a sum of one
        across all bins.

        Arguments:
            frame_img: The input image in BGR color space, assumed to have shape
                (height, width, 3) where the last dimension represents the BGR channels.
            bins: The number of bins to use for the histogram.
            normalize: A boolean flag that determines whether the histogram should be
                normalized such that the sum of all histogram bins equals 1.

        Returns:
            A 1D numpy array of length equal to `bins`, representing the histogram of the
            luma channel. Each element in the array represents the count (or frequency) of
            a particular luma value in the image. If normalized, these values represent the
            relative frequency.

        Example:

        .. code:: python

            img = cv2.imread("path_to_image.jpg")
            hist = HistogramDetector.calculate_histogram(img, bins=256, normalize=True)
            assert hist.shape == (256,)
        """
        # Extract Luma channel from the frame image
        y, _, _ = cv2.split(cv2.cvtColor(frame_img, cv2.COLOR_BGR2YUV))

        # Create the histogram with a bin for every rgb value
        hist = cv2.calcHist([y], [0], None, [bins], [0, 256])

        if normalize:
            # Normalize the histogram
            hist = cv2.normalize(hist, hist).flatten()

        return hist

    def get_metrics(self) -> list[str]:
        return [self._metric_key]



#
#            PySceneDetect: Python-Based Video Scene Detector
#   -------------------------------------------------------------------
#     [  Site:    https://scenedetect.com                           ]
#     [  Docs:    https://scenedetect.com/docs/                     ]
#     [  Github:  https://github.com/Breakthrough/PySceneDetect/    ]
#
# Copyright (C) 2026 Brandon Castellano <http://www.bcastell.com>.
# PySceneDetect is licensed under the BSD 3-Clause License; see the
# included LICENSE file, or visit one of the above pages for details.
#
"""Codec / Container / Geometry

Verifies that PySceneDetect can handle various codecs, containers, and video properties.
"""

import subprocess

import pytest

from scenedetect import ContentDetector, SceneManager, open_video

MATRIX = [
    ("libx264", "mp4", []),
    ("libx265", "mkv", []),
    ("libvpx-vp9", "webm", []),
    ("libx264", "mp4", ["-vf", "transpose=1"]),  # Rotation
    ("libx264", "mp4", ["-pix_fmt", "yuv400p"]),  # Grayscale
    ("libx264", "mp4", ["-vf", "scale=3840:2160"]),  # 4K UHD
    ("libx264", "mp4", ["-vf", "fps=120"]),  # 120 fps high frame rate
    ("libx265", "mp4", ["-pix_fmt", "yuv420p10le"]),  # 10-bit HEVC (HDR-adjacent)
]


@pytest.mark.release
@pytest.mark.parametrize("codec, container, extra_args", MATRIX)
@pytest.mark.parametrize("backend", ["opencv", "pyav"])
def test_synthetic_matrix(synthetic_matrix_generator, codec, container, extra_args, backend):
    try:
        video_path = synthetic_matrix_generator(codec, container, extra_args)
    except subprocess.CalledProcessError:
        pytest.skip(f"Codec {codec} or container {container} not supported by ffmpeg.")

    video = open_video(video_path, backend=backend)
    scene_manager = SceneManager()
    scene_manager.add_detector(ContentDetector())
    scene_manager.detect_scenes(video)

    # Ensure it processed some frames
    assert video.frame_number > 0
    # Plausible duration
    assert video.duration is not None
    assert abs(video.duration.seconds - 2.0) < 0.2



#
#            PySceneDetect: Python-Based Video Scene Detector
#   -------------------------------------------------------------------
#     [  Site:    https://scenedetect.com                           ]
#     [  Docs:    https://scenedetect.com/docs/                     ]
#     [  Github:  https://github.com/Breakthrough/PySceneDetect/    ]
#
# Copyright (C) 2018 Brandon Castellano <http://www.bcastell.com>.
# PySceneDetect is licensed under the BSD 3-Clause License; see the
# included LICENSE file, or visit one of the above pages for details.
#
"""DEPRECATED"""

import warnings

warnings.warn(
    "The `frame_timecode` submodule is deprecated, import from the base package instead.",
    DeprecationWarning,
    stacklevel=2,
)

from scenedetect.common import *  # noqa: E402, F403



#
#            PySceneDetect: Python-Based Video Scene Detector
#   -------------------------------------------------------------------
#     [  Site:    https://scenedetect.com                           ]
#     [  Docs:    https://scenedetect.com/docs/                     ]
#     [  Github:  https://github.com/Breakthrough/PySceneDetect/    ]
#
# Copyright (C) 2026 Brandon Castellano <http://www.bcastell.com>.
# PySceneDetect is licensed under the BSD 3-Clause License; see the
# included LICENSE file, or visit one of the above pages for details.
#
"""CLI Permutation Smoke Tests

Exercises CLI command chains via subprocess.
"""

import os
import subprocess
import sys

import pytest


def _run(args, cwd):
    result = subprocess.run(
        [sys.executable, "-m", "scenedetect", *args],
        cwd=cwd,
        capture_output=True,
        text=True,
    )
    return result


@pytest.mark.release
def test_cli_chain_smoke(test_video_file, tmp_path):
    # detect-content save-images list-scenes chain.
    result = _run(
        [
            "-i",
            os.path.abspath(test_video_file),
            "-o",
            str(tmp_path),
            "detect-content",
            "save-images",
            "list-scenes",
        ],
        cwd=os.path.abspath(os.path.dirname(test_video_file) + "/../.."),
    )
    assert result.returncode == 0, f"stderr:\n{result.stderr}\nstdout:\n{result.stdout}"
    csvs = [p for p in tmp_path.iterdir() if p.suffix == ".csv"]
    images = [p for p in tmp_path.iterdir() if p.suffix == ".jpg"]
    assert csvs, "No scenes CSV produced"
    assert images, "No scene images produced"


@pytest.mark.release
def test_cli_range_smoke(test_video_file, tmp_path):
    result = _run(
        [
            "-i",
            os.path.abspath(test_video_file),
            "-o",
            str(tmp_path),
            "time",
            "-e",
            "2s",
            "detect-content",
            "list-scenes",
        ],
        cwd=os.path.abspath(os.path.dirname(test_video_file) + "/../.."),
    )
    assert result.returncode == 0, f"stderr:\n{result.stderr}\nstdout:\n{result.stdout}"


@pytest.mark.release
def test_cli_stats_roundtrip(test_video_file, tmp_path):
    stats_path = tmp_path / "stats.csv"
    repo_cwd = os.path.abspath(os.path.dirname(test_video_file) + "/../..")

    # First run: generate stats.
    run1 = _run(
        [
            "-i",
            os.path.abspath(test_video_file),
            "-s",
            str(stats_path),
            "-o",
            str(tmp_path),
            "detect-content",
            "list-scenes",
            "-f",
            "run1",
        ],
        cwd=repo_cwd,
    )
    assert run1.returncode == 0, run1.stderr
    assert stats_path.exists()

    # Second run: reuse stats.
    run2 = _run(
        [
            "-i",
            os.path.abspath(test_video_file),
            "-s",
            str(stats_path),
            "-o",
            str(tmp_path),
            "detect-content",
            "list-scenes",
            "-f",
            "run2",
        ],
        cwd=repo_cwd,
    )
    assert run2.returncode == 0, run2.stderr

    def _cuts(csv_path):
        # First line is the cut-list summary; extract it for comparison.
        return csv_path.read_text().splitlines()[0]

    assert _cuts(tmp_path / "run1.csv") == _cuts(tmp_path / "run2.csv"), (
        "Cut list differs between stats-producing run and stats-consuming run."
    )


@pytest.mark.release
def test_cli_min_scene_len_smoke(test_video_file, tmp_path):
    # A min-scene-len longer than the video collapses everything to a single scene.
    result = _run(
        [
            "-i",
            os.path.abspath(test_video_file),
            "-o",
            str(tmp_path),
            "detect-content",
            "--min-scene-len",
            "1000s",
            "list-scenes",
        ],
        cwd=os.path.abspath(os.path.dirname(test_video_file) + "/../.."),
    )
    assert result.returncode == 0, f"stderr:\n{result.stderr}\nstdout:\n{result.stdout}"


@pytest.mark.release
def test_cli_save_fcp_smoke(test_video_file, tmp_path):
    """save-fcp writes a well-formed Final Cut Pro XML."""
    import xml.etree.ElementTree as ET

    result = _run(
        [
            "-i",
            os.path.abspath(test_video_file),
            "-o",
            str(tmp_path),
            "detect-content",
            "save-fcp",
        ],
        cwd=os.path.abspath(os.path.dirname(test_video_file) + "/../.."),
    )
    assert result.returncode == 0, f"stderr:\n{result.stderr}\nstdout:\n{result.stdout}"
    xml_files = [p for p in tmp_path.iterdir() if p.suffix == ".xml"]
    assert xml_files, "save-fcp produced no .xml file"
    # Parse must succeed; root <fcpxml> or <xmeml> depending on the FCP variant.
    root = ET.parse(xml_files[0]).getroot()
    assert root.tag in ("fcpxml", "xmeml"), f"Unexpected root element: {root.tag}"


@pytest.mark.release
def test_cli_save_qp_smoke(test_video_file, tmp_path):
    """save-qp writes a QP file with `<frame> I <shift>` lines for scene boundaries."""
    result = _run(
        [
            "-i",
            os.path.abspath(test_video_file),
            "-o",
            str(tmp_path),
            "detect-content",
            "save-qp",
        ],
        cwd=os.path.abspath(os.path.dirname(test_video_file) + "/../.."),
    )
    assert result.returncode == 0, f"stderr:\n{result.stderr}\nstdout:\n{result.stdout}"
    qp_files = [p for p in tmp_path.iterdir() if p.suffix == ".qp"]
    assert qp_files, "save-qp produced no .qp file"
    contents = qp_files[0].read_text().strip()
    assert contents, "save-qp produced an empty file"
    # Each line must be `<frame_number> I <shift>` where shift is an integer.
    for line in contents.splitlines():
        parts = line.split()
        assert len(parts) == 3 and parts[0].isdigit() and parts[1] == "I", (
            f"Malformed QP line: {line!r}"
        )
        int(parts[2])  # shift must parse as int (may be negative)


@pytest.mark.release
def test_cli_save_html_smoke(test_video_file, tmp_path):
    """save-html replaces the deprecated export-html and produces an HTML report.

    Note: save-html lacks its own --output option and ignores the global -o, so the
    file is routed via --filename with an absolute path.
    """
    out_html = tmp_path / "scenes.html"
    result = _run(
        [
            "-i",
            os.path.abspath(test_video_file),
            "detect-content",
            "save-html",
            "--filename",
            str(out_html),
            "--no-images",
        ],
        cwd=os.path.abspath(os.path.dirname(test_video_file) + "/../.."),
    )
    assert result.returncode == 0, f"stderr:\n{result.stderr}\nstdout:\n{result.stdout}"
    assert out_html.exists(), f"save-html produced no file at {out_html}"
    contents = out_html.read_text(encoding="utf-8")
    # The output is an HTML fragment (a <table> of scenes), not a full document.
    lowered = contents.lower()
    assert "<table" in lowered and "</table>" in lowered, (
        f"save-html output is missing the scenes <table>:\n{contents[:500]}"
    )


@pytest.mark.release
def test_cli_save_edl_start_timecode_smoke(test_video_file, tmp_path):
    """save-edl --start-timecode produces an EDL where event timestamps are offset by the
    requested start. Both SMPTE (HH:MM:SS:FF) and 8-digit (HHMMSSFF) inputs must be accepted."""
    repo_cwd = os.path.abspath(os.path.dirname(test_video_file) + "/../..")

    def _edl(form: str, out_dir):
        result = _run(
            [
                "-i",
                os.path.abspath(test_video_file),
                "-o",
                str(out_dir),
                "detect-content",
                "save-edl",
                "--start-timecode",
                form,
            ],
            cwd=repo_cwd,
        )
        assert result.returncode == 0, (
            f"start-timecode {form!r} failed:\nstderr:\n{result.stderr}\nstdout:\n{result.stdout}"
        )
        edls = [p for p in out_dir.iterdir() if p.suffix == ".edl"]
        assert edls, f"save-edl --start-timecode {form!r} produced no .edl file"
        return edls[0].read_text()

    # SMPTE form.
    smpte_dir = tmp_path / "smpte"
    smpte_dir.mkdir()
    smpte_text = _edl("01:00:00:00", smpte_dir)
    # 8-digit form (semantically equivalent to 01:00:00:00).
    digit_dir = tmp_path / "digit"
    digit_dir.mkdir()
    digit_text = _edl("01000000", digit_dir)
    # Both EDLs must contain at least one event line with the 01:00:... offset visible.
    assert "01:00:" in smpte_text, f"SMPTE start TC not propagated to EDL:\n{smpte_text}"
    assert "01:00:" in digit_text, f"8-digit start TC not propagated to EDL:\n{digit_text}"



#
#            PySceneDetect: Python-Based Video Scene Detector
#   -------------------------------------------------------------------
#     [  Site:    https://scenedetect.com                           ]
#     [  Docs:    https://scenedetect.com/docs/                     ]
#     [  Github:  https://github.com/Breakthrough/PySceneDetect/    ]
#
# Copyright (C) 2026 Brandon Castellano <http://www.bcastell.com>.
# PySceneDetect is licensed under the BSD 3-Clause License; see the
# included LICENSE file, or visit one of the above pages for details.
#
"""Benchmark dataset definitions and registry.

Each :class:`Dataset` is a corpus of :class:`Sample` records (video file + typed ground truth)
loaded eagerly at construction. Ground-truth files for the supported corpora are at most a few
hundred kilobytes total, so eager loading avoids re-reading the same files for every sweep cell.

Add a new dataset by:

1. Subclassing :class:`Dataset` and populating ``self._samples`` in ``__init__``.
2. Registering it in :data:`DATASETS` under the name used by ``--dataset``.
"""

from __future__ import annotations

import glob
import json
import logging
import os
from collections.abc import Iterator
from dataclasses import dataclass
from pathlib import Path

from benchmark.evaluator import EventInterval, Frames, GroundTruth

logger = logging.getLogger("pyscenedetect")


@dataclass(frozen=True)
class Sample:
    """One scored video: a path on disk plus its typed ground truth."""

    video_file: Path
    ground_truth: GroundTruth


class Dataset:
    """Iterable corpus of :class:`Sample` records.

    Subclasses populate ``self._samples`` in their constructor; this base provides the iteration
    and length protocol. ``event_types`` advertises which TRECVID-SBD event categories the
    dataset's ground truth contains, so consumers can skip columns/tables for categories that
    have no events (e.g. fade transitions on BBC/AutoShot).
    """

    event_types: frozenset[str] = frozenset({"hard_cut"})
    _samples: list[Sample]

    def __iter__(self) -> Iterator[Sample]:
        return iter(self._samples)

    def __len__(self) -> int:
        return len(self._samples)


def _read_tab_separated_cuts(scene_file: str) -> list[Frames]:
    """Parse a BBC/AutoShot-style annotation file.

    Each line is tab-separated; the second column is the 0-based frame index of a
    hard cut. Returns 1-based frame indices, matching the convention used by
    :class:`scenedetect.FrameTimecode`.
    """
    with open(scene_file) as f:
        return [int(line.strip().split("\t")[1]) + 1 for line in f]


class BBCDataset(Dataset):
    """The BBC Planet Earth dataset.

    Baraldi et al., "A Deep Siamese Network for Scene Detection in Broadcast Videos",
    ACM Multimedia 2015. https://arxiv.org/abs/1510.08893

    11 long-form videos (``BBC/videos/bbc_<id>.mp4``) with hard-cut annotations in
    ``BBC/fixed/<id>-scenes.txt``.
    """

    def __init__(self, dataset_dir: str):
        video_files = sorted(glob.glob(os.path.join(dataset_dir, "videos", "*.mp4")))
        scene_files = sorted(glob.glob(os.path.join(dataset_dir, "fixed", "*.txt")))
        if len(video_files) != len(scene_files):
            raise ValueError(
                f"BBC dataset at {dataset_dir!r}: {len(video_files)} videos but "
                f"{len(scene_files)} annotation files."
            )
        self._samples: list[Sample] = []
        for video_file, scene_file in zip(video_files, scene_files, strict=True):
            video_id = os.path.basename(video_file).replace("bbc_", "").split(".")[0]
            scene_id = os.path.basename(scene_file).split("-")[0]
            if video_id != scene_id:
                raise ValueError(f"BBC id mismatch: {video_file} vs {scene_file}")
            self._samples.append(
                Sample(
                    video_file=Path(video_file),
                    ground_truth=GroundTruth(hard_cuts=_read_tab_separated_cuts(scene_file)),
                )
            )


class AutoShotDataset(Dataset):
    """The AutoShot dataset (test splits).

    Zhu et al., "AutoShot: A Short Video Dataset and State-of-the-Art Shot Boundary
    Detection", CVPRW 2023. The original test set has 200 videos; 36 are no longer
    publicly available, so the corpus iterates over whatever is present on disk.

    Videos at ``AutoShot/videos/<id>.mp4``, hard-cut annotations at
    ``AutoShot/annotations/<id>.txt``.
    """

    def __init__(self, dataset_dir: str):
        # 36 of the original 200 videos are no longer publicly available, so intersect
        # by id rather than zipping the directory listings strictly.
        videos_by_id = {
            os.path.basename(p).split(".")[0]: p
            for p in glob.glob(os.path.join(dataset_dir, "videos", "*.mp4"))
        }
        scenes_by_id = {
            os.path.basename(p).split(".")[0]: p
            for p in glob.glob(os.path.join(dataset_dir, "annotations", "*.txt"))
        }
        self._samples: list[Sample] = [
            Sample(
                video_file=Path(videos_by_id[vid]),
                ground_truth=GroundTruth(hard_cuts=_read_tab_separated_cuts(scenes_by_id[vid])),
            )
            for vid in sorted(videos_by_id.keys() & scenes_by_id.keys())
        ]


class ClipShotsDataset(Dataset):
    """The ClipShots dataset (test split by default).

    Tang et al., "Fast Video Shot Transition Localization with Deep Structured Models",
    ACCV 2018. https://github.com/Tangshitao/ClipShots

    The only in-tree dataset with typed gradual-transition (fade/dissolve) ground truth in
    addition to hard cuts. Layout under ``ClipShots/``::

        annotations/{train,test,only_gradual}.json
        video_lists/{train,test,only_gradual}.txt   (optional split filter)
        videos/*.mp4

    Each annotation entry is ``{"transitions": [[start, end], ...], "frame_num": float}``.
    Hard cuts are single-frame spans (``end == start + 1``); wider spans are gradual
    transitions. Unlike the BBC/AutoShot annotations, ClipShots frame indices already match
    PySceneDetect's boundary-frame convention (the prediction's ``frame_num`` lines up with
    ``transition[1]`` directly), so no offset is applied here.

    Loading rules:

    - Videos listed in ``video_lists/<split>.txt`` but absent from the annotations JSON are
      silently ignored (the filter runs against the JSON, not the other way).
    - Annotations whose ``.mp4`` is not on disk are skipped (so partial corpora work).
    - Malformed transitions (fewer than 2 entries, negative span, zero-width span) are
      skipped with a warning rather than crashing the load.

    Only the ``ClipShotsDataset(dir, split=...)`` constructor honors a non-default split;
    the registry entry in :data:`DATASETS` always loads the ``test`` split.
    """

    event_types = frozenset({"hard_cut", "fade"})

    def __init__(self, dataset_dir: str, split: str = "test"):
        ann_path = os.path.join(dataset_dir, "annotations", f"{split}.json")
        videos_dir = os.path.join(dataset_dir, "videos")
        with open(ann_path) as f:
            annotations: dict = json.load(f)
        split_list_path = os.path.join(dataset_dir, "video_lists", f"{split}.txt")
        if os.path.exists(split_list_path):
            with open(split_list_path) as allow_f:
                allowed = {line.strip() for line in allow_f if line.strip()}
            annotations = {k: v for k, v in annotations.items() if k in allowed}
        total = len(annotations)
        skipped_missing = 0
        self._samples: list[Sample] = []
        for video_name in sorted(annotations):
            video_path = os.path.join(videos_dir, video_name)
            if not os.path.exists(video_path):
                skipped_missing += 1
                continue
            hard_cuts: list[Frames] = []
            fades: list[EventInterval] = []
            # `... or []` (not `.get(k, [])`) so an explicit JSON `null` is treated as empty.
            for transition in annotations[video_name].get("transitions") or []:
                if len(transition) < 2:
                    logger.warning("ClipShots %s: malformed transition %r", video_name, transition)
                    continue
                start, end = int(transition[0]), int(transition[1])
                span = end - start
                if span == 1:
                    hard_cuts.append(end)
                elif span > 1:
                    fades.append(EventInterval(start=start, end=end))
                else:
                    logger.warning(
                        "ClipShots %s: skipping degenerate transition %r", video_name, transition
                    )
            self._samples.append(
                Sample(
                    video_file=Path(video_path),
                    ground_truth=GroundTruth(hard_cuts=hard_cuts, fades=fades),
                )
            )
        logger.info(
            "ClipShots %s: loaded %d/%d samples (%d videos missing on disk)",
            split,
            len(self._samples),
            total,
            skipped_missing,
        )


# Mapping of --dataset names to constructors. Typed as a plain callable so
# subclass-specific positional signatures (each takes ``dataset_dir: str``)
# aren't widened away by the base ``Dataset`` class's empty ``__init__``.
DATASETS: dict[str, type] = {
    "BBC": BBCDataset,
    "AutoShot": AutoShotDataset,
    "ClipShots": ClipShotsDataset,
}


def resolve_dataset(name: str, root: str | None) -> Dataset:
    """Instantiate the named dataset.

    ``root`` overrides the default repo-relative path; pass ``None`` (or the empty string)
    to use ``benchmark/<name>/``.
    """
    base = root if root else "benchmark"
    return DATASETS[name](os.path.join(base, name))



#
#            PySceneDetect: Python-Based Video Scene Detector
#   -------------------------------------------------------------------
#     [  Site:    https://scenedetect.com                           ]
#     [  Docs:    https://scenedetect.com/docs/                     ]
#     [  Github:  https://github.com/Breakthrough/PySceneDetect/    ]
#
# Copyright (C) 2026 Brandon Castellano <http://www.bcastell.com>.
# PySceneDetect is licensed under the BSD 3-Clause License; see the
# included LICENSE file, or visit one of the above pages for details.
#
"""Shared test helpers."""

from click.testing import CliRunner

from scenedetect._cli import scenedetect as _scenedetect_cli
from scenedetect._cli.context import CliContext
from scenedetect._cli.controller import run_scenedetect


def invoke_cli(args: list[str], catch_exceptions: bool = False) -> tuple[int, str]:
    """Invoke the scenedetect CLI in-process using Click's CliRunner.

    Replicates the two-step execution of ``__main__.py``:

    1. ``scenedetect.main(obj=context)`` - parse args and register callbacks on ``CliContext``
    2. ``run_scenedetect(context)`` - execute detection and output commands

    Returns ``(exit_code, output_text)``.
    """
    context = CliContext()
    runner = CliRunner()
    result = runner.invoke(_scenedetect_cli, args, obj=context, catch_exceptions=catch_exceptions)
    if result.exit_code == 0:
        run_scenedetect(context)
    return result.exit_code, result.output
